# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi group` commands"""
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import echo
from aiida.cmdline.utils.decorators import with_dbenv
from aiida.common.exceptions import UniquenessError
from aiida.common.links import GraphTraversalRules


@verdi.group('group')
def verdi_group():
    """Create, inspect and manage groups of nodes."""


@verdi_group.command('add-nodes')
@options.GROUP(required=True)
@options.FORCE()
@arguments.NODES()
@with_dbenv()
def group_add_nodes(group, force, nodes):
    """Add nodes to a group."""
    if not force:
        click.confirm(f'Do you really want to add {len(nodes)} nodes to {group}?', abort=True)

    group.add_nodes(nodes)


@verdi_group.command('remove-nodes')
@options.GROUP(required=True)
@arguments.NODES()
@options.GROUP_CLEAR()
@options.FORCE()
@with_dbenv()
def group_remove_nodes(group, nodes, clear, force):
    """Remove nodes from a group."""
    from aiida.orm import Group, Node, QueryBuilder

    if nodes and clear:
        echo.echo_critical(
            'Specify either the `--clear` flag to remove all nodes or the identifiers of the nodes you want to remove.'
        )

    if not force:

        if nodes:
            node_pks = [node.pk for node in nodes]

            query = QueryBuilder()
            query.append(Group, filters={'id': group.pk}, tag='group')
            query.append(Node, with_group='group', filters={'id': {'in': node_pks}}, project='id')

            group_node_pks = query.all(flat=True)

            if not group_node_pks:
                echo.echo_critical(f'None of the specified nodes are in {group}.')

            if len(node_pks) > len(group_node_pks):
                node_pks = set(node_pks).difference(set(group_node_pks))
                echo.echo_warning(f'{len(node_pks)} nodes with PK {node_pks} are not in {group}.')

            message = f'Are you sure you want to remove {len(group_node_pks)} nodes from {group}?'

        elif clear:
            message = f'Are you sure you want to remove ALL the nodes from {group}?'
        else:
            echo.echo_critical(f'No nodes were provided for removal from {group}.')

        click.confirm(message, abort=True)

    if clear:
        group.clear()
    else:
        group.remove_nodes(nodes)


@verdi_group.command('move-nodes')
@arguments.NODES()
@click.option('-s', '--source-group', type=types.GroupParamType(), required=True, help='The group whose nodes to move.')
@click.option(
    '-t', '--target-group', type=types.GroupParamType(), required=True, help='The group to which the nodes are moved.'
)
@options.FORCE(help='Do not ask for confirmation and skip all checks.')
@options.ALL(help='Move all nodes from the source to the target group.')
@with_dbenv()
def group_move_nodes(source_group, target_group, force, nodes, all_entries):
    """Move the specified NODES from one group to another."""
    from aiida.orm import Group, Node, QueryBuilder

    if source_group.pk == target_group.pk:
        echo.echo_critical(f'Source and target group are the same: {source_group}.')

    if not nodes:
        if all_entries:
            nodes = list(source_group.nodes)
        else:
            echo.echo_critical('Neither NODES or the `-a, --all` option was specified.')

    node_pks = [node.pk for node in nodes]

    if not all_entries:
        query = QueryBuilder()
        query.append(Group, filters={'id': source_group.pk}, tag='group')
        query.append(Node, with_group='group', filters={'id': {'in': node_pks}}, project='id')

        source_group_node_pks = query.all(flat=True)

        if not source_group_node_pks:
            echo.echo_critical(f'None of the specified nodes are in {source_group}.')

        if len(node_pks) > len(source_group_node_pks):
            absent_node_pks = set(node_pks).difference(set(source_group_node_pks))
            echo.echo_warning(f'{len(absent_node_pks)} nodes with PK {absent_node_pks} are not in {source_group}.')
            nodes = [node for node in nodes if node.pk in source_group_node_pks]
            node_pks = set(node_pks).difference(absent_node_pks)

    query = QueryBuilder()
    query.append(Group, filters={'id': target_group.pk}, tag='group')
    query.append(Node, with_group='group', filters={'id': {'in': node_pks}}, project='id')

    target_group_node_pks = query.all(flat=True)

    if target_group_node_pks:
        echo.echo_warning(
            f'{len(target_group_node_pks)} nodes with PK {set(target_group_node_pks)} are already in '
            f'{target_group}. These will still be removed from {source_group}.'
        )

    if not force:
        click.confirm(
            f'Are you sure you want to move {len(nodes)} nodes from {source_group} '
            f'to {target_group}?', abort=True
        )

    source_group.remove_nodes(nodes)
    target_group.add_nodes(nodes)


@verdi_group.command('delete')
@arguments.GROUP()
@options.FORCE()
@click.option(
    '--delete-nodes', is_flag=True, default=False, help='Delete all nodes in the group along with the group itself.'
)
@options.graph_traversal_rules(GraphTraversalRules.DELETE.value)
@options.DRY_RUN()
@with_dbenv()
def group_delete(group, delete_nodes, dry_run, force, **traversal_rules):
    """Delete a group and (optionally) the nodes it contains."""
    from aiida import orm
    from aiida.tools import delete_group_nodes

    if not (force or dry_run):
        click.confirm(f'Are you sure you want to delete {group}?', abort=True)
    elif dry_run:
        echo.echo_report(f'Would have deleted {group}.')

    if delete_nodes:

        def _dry_run_callback(pks):
            if not pks or force:
                return False
            echo.echo_warning(f'YOU ARE ABOUT TO DELETE {len(pks)} NODES! THIS CANNOT BE UNDONE!')
            return not click.confirm('Do you want to continue?', abort=True)

        _, nodes_deleted = delete_group_nodes([group.pk], dry_run=dry_run or _dry_run_callback, **traversal_rules)
        if not nodes_deleted:
            # don't delete the group if the nodes were not deleted
            return

    if not dry_run:
        group_str = str(group)
        orm.Group.collection.delete(group.pk)
        echo.echo_success(f'{group_str} deleted.')


@verdi_group.command('relabel')
@arguments.GROUP()
@click.argument('label', type=click.STRING)
@with_dbenv()
def group_relabel(group, label):
    """Change the label of a group."""
    try:
        group.label = label
    except UniquenessError as exception:
        echo.echo_critical(str(exception))
    else:
        echo.echo_success(f"Label changed to '{label}'")


@verdi_group.command('description')
@arguments.GROUP()
@click.argument('description', type=click.STRING, required=False)
@with_dbenv()
def group_description(group, description):
    """Change the description of a group.

    If no description is defined, the current description will simply be echoed.
    """
    if description:
        group.description = description
        echo.echo_success(f'Changed the description of {group}.')
    else:
        echo.echo(group.description)


@verdi_group.command('show')
@options.RAW(help='Show only a space-separated list of PKs of the calculations in the group')
@options.LIMIT()
@click.option(
    '-u',
    '--uuid',
    is_flag=True,
    default=False,
    help='Show UUIDs together with PKs. Note: if the --raw option is also passed, PKs are not printed, but only UUIDs.'
)
@arguments.GROUP()
@with_dbenv()
def group_show(group, raw, limit, uuid):
    """Show information for a given group."""
    from tabulate import tabulate

    from aiida.common import timezone
    from aiida.common.utils import str_timedelta

    if limit:
        node_iterator = group.nodes[:limit]
    else:
        node_iterator = group.nodes

    if raw:
        if uuid:
            echo.echo(' '.join(str(_.uuid) for _ in node_iterator))
        else:
            echo.echo(' '.join(str(_.pk) for _ in node_iterator))
    else:
        type_string = group.type_string
        desc = group.description
        now = timezone.now()

        table = []
        table.append(['Group label', group.label])
        table.append(['Group type_string', type_string])
        table.append(['Group description', desc if desc else '<no description>'])
        echo.echo(tabulate(table))

        table = []
        header = []
        if uuid:
            header.append('UUID')
        header.extend(['PK', 'Type', 'Created'])
        echo.echo('# Nodes:')
        for node in node_iterator:
            row = []
            if uuid:
                row.append(node.uuid)
            row.append(node.pk)
            row.append(node.node_type.rsplit('.', 2)[1])
            row.append(str_timedelta(now - node.ctime, short=True, negative_to_zero=True))
            table.append(row)
        echo.echo(tabulate(table, headers=header))


@verdi_group.command('list')
@options.ALL_USERS(help='Show groups for all users, rather than only for the current user.')
@options.USER(help='Add a filter to show only groups belonging to a specific user.')
@options.ALL(help='Show groups of all types.')
@options.TYPE_STRING()
@click.option(
    '-d',
    '--with-description',
    'with_description',
    is_flag=True,
    default=False,
    help='Show also the group description.'
)
@click.option('-C', '--count', is_flag=True, default=False, help='Show also the number of nodes in the group.')
@options.PAST_DAYS(help='Add a filter to show only groups created in the past N days.', default=None)
@click.option(
    '-s',
    '--startswith',
    type=click.STRING,
    default=None,
    help='Add a filter to show only groups for which the label begins with STRING.'
)
@click.option(
    '-e',
    '--endswith',
    type=click.STRING,
    default=None,
    help='Add a filter to show only groups for which the label ends with STRING.'
)
@click.option(
    '-c',
    '--contains',
    type=click.STRING,
    default=None,
    help='Add a filter to show only groups for which the label contains STRING.'
)
@options.ORDER_BY(type=click.Choice(['id', 'label', 'ctime']), default='id')
@options.ORDER_DIRECTION()
@options.NODE(help='Show only the groups that contain the node.')
@with_dbenv()
def group_list(
    all_users, user, all_entries, type_string, with_description, count, past_days, startswith, endswith, contains,
    order_by, order_dir, node
):
    """Show a list of existing groups."""
    # pylint: disable=too-many-branches,too-many-arguments,too-many-locals,too-many-statements
    import datetime

    from tabulate import tabulate

    from aiida import orm
    from aiida.common import timezone
    from aiida.common.escaping import escape_for_sql_like

    builder = orm.QueryBuilder()
    filters = {}

    # Have to specify the default for `type_string` here instead of directly in the option otherwise it will always
    # raise above if the user specifies just the `--group-type` option. Once that option is removed, the default can
    # be moved to the option itself.
    if type_string is None:
        type_string = 'core'

    if not all_entries:
        if '%' in type_string or '_' in type_string:
            filters['type_string'] = {'like': type_string}
        else:
            filters['type_string'] = type_string

    # Creation time
    if past_days:
        filters['time'] = {'>': timezone.now() - datetime.timedelta(days=past_days)}

    # Query for specific group labels
    filters['or'] = []
    if startswith:
        filters['or'].append({'label': {'like': f'{escape_for_sql_like(startswith)}%'}})
    if endswith:
        filters['or'].append({'label': {'like': f'%{escape_for_sql_like(endswith)}'}})
    if contains:
        filters['or'].append({'label': {'like': f'%{escape_for_sql_like(contains)}%'}})

    builder.append(orm.Group, filters=filters, tag='group', project='*')

    # Query groups that belong to specific user
    if user:
        user_email = user.email
    else:
        # By default: only groups of this user
        user_email = orm.User.collection.get_default().email

    # Query groups that belong to all users
    if not all_users:
        builder.append(orm.User, filters={'email': {'==': user_email}}, with_group='group')

    # Query groups that contain a particular node
    if node:
        builder.append(orm.Node, filters={'id': {'==': node.pk}}, with_group='group')

    builder.order_by({orm.Group: {order_by: order_dir}})
    result = builder.all()

    projection_lambdas = {
        'pk': lambda group: str(group.pk),
        'label': lambda group: group.label,
        'type_string': lambda group: group.type_string,
        'count': lambda group: group.count(),
        'user': lambda group: group.user.email.strip(),
        'description': lambda group: group.description
    }

    table = []
    projection_header = ['PK', 'Label', 'Type string', 'User']
    projection_fields = ['pk', 'label', 'type_string', 'user']

    if with_description:
        projection_header.append('Description')
        projection_fields.append('description')

    if count:
        projection_header.append('Node count')
        projection_fields.append('count')

    for group in result:
        table.append([projection_lambdas[field](group[0]) for field in projection_fields])

    if not all_entries:
        echo.echo_report('To show groups of all types, use the `-a/--all` option.')

    if not table:
        echo.echo_report('No groups found matching the specified criteria.')
    else:
        echo.echo(tabulate(table, headers=projection_header))


@verdi_group.command('create')
@click.argument('group_label', nargs=1, type=click.STRING)
@with_dbenv()
def group_create(group_label):
    """Create an empty group with a given label."""
    from aiida import orm

    group, created = orm.Group.collection.get_or_create(label=group_label)

    if created:
        echo.echo_success(f"Group created with PK = {group.pk} and label '{group.label}'.")
    else:
        echo.echo_report(f"Group with label '{group.label}' already exists: {group}.")


@verdi_group.command('copy')
@arguments.GROUP('source_group')
@click.argument('destination_group', nargs=1, type=click.STRING)
@with_dbenv()
def group_copy(source_group, destination_group):
    """Duplicate a group.

    More in detail, add all nodes from the source group to the destination group.
    Note that the destination group may not exist."""
    from aiida import orm

    dest_group, created = orm.Group.collection.get_or_create(label=destination_group)

    # Issue warning if destination group is not empty and get user confirmation to continue
    if not created and not dest_group.is_empty:
        echo.echo_warning(f'Destination {dest_group} already exists and is not empty.')
        click.confirm('Do you wish to continue anyway?', abort=True)

    # Copy nodes
    dest_group.add_nodes(list(source_group.nodes))
    echo.echo_success(f'Nodes copied from {source_group} to {dest_group}.')


@verdi_group.group('path')
def verdi_group_path():
    """Inspect groups of nodes, with delimited label paths."""


@verdi_group_path.command('ls')
@click.argument('path', type=click.STRING, required=False)
@options.TYPE_STRING(default='core', help='Filter to only include groups of this type string.')
@click.option('-R', '--recursive', is_flag=True, default=False, help='Recursively list sub-paths encountered.')
@click.option('-l', '--long', 'as_table', is_flag=True, default=False, help='List as a table, with sub-group count.')
@click.option(
    '-d',
    '--with-description',
    'with_description',
    is_flag=True,
    default=False,
    help='Show also the group description.'
)
@click.option(
    '--no-virtual',
    'no_virtual',
    is_flag=True,
    default=False,
    help='Only show paths that fully correspond to an existing group.'
)
@click.option('--no-warn', is_flag=True, default=False, help='Do not issue a warning if any paths are invalid.')
@with_dbenv()
def group_path_ls(path, type_string, recursive, as_table, no_virtual, with_description, no_warn):
    # pylint: disable=too-many-arguments,too-many-branches
    """Show a list of existing group paths."""
    from aiida.plugins import GroupFactory
    from aiida.tools.groups.paths import GroupPath, InvalidPath

    try:
        path = GroupPath(path or '', cls=GroupFactory(type_string), warn_invalid_child=not no_warn)
    except InvalidPath as err:
        echo.echo_critical(str(err))

    if recursive:
        children = path.walk()
    else:
        children = path.children

    if as_table or with_description:
        from tabulate import tabulate
        headers = ['Path', 'Sub-Groups']
        if with_description:
            headers.append('Description')
        rows = []
        for child in sorted(children):
            if no_virtual and child.is_virtual:
                continue
            row = [
                child.path if child.is_virtual else click.style(child.path, bold=True),
                len([c for c in child.walk() if not c.is_virtual])
            ]
            if with_description:
                row.append('-' if child.is_virtual else child.get_group().description)
            rows.append(row)
        echo.echo(tabulate(rows, headers=headers))
    else:
        for child in sorted(children):
            if no_virtual and child.is_virtual:
                continue
            echo.echo(child.path, bold=not child.is_virtual)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Function that starts a daemon runner."""
import asyncio
import logging
import signal

from aiida.common.log import configure_logging
from aiida.engine.daemon.client import get_daemon_client
from aiida.engine.runners import Runner
from aiida.manage import get_manager

LOGGER = logging.getLogger(__name__)


async def shutdown_runner(runner: Runner) -> None:
    """Cleanup tasks tied to the service's shutdown."""
    from asyncio import all_tasks, current_task

    LOGGER.info('Received signal to shut down the daemon runner')
    tasks = [task for task in all_tasks() if task is not current_task()]

    for task in tasks:
        task.cancel()

    await asyncio.gather(*tasks, return_exceptions=True)
    runner.close()

    LOGGER.info('Daemon runner stopped')


def start_daemon() -> None:
    """Start a daemon runner for the currently configured profile."""
    daemon_client = get_daemon_client()
    configure_logging(daemon=True, daemon_log_file=daemon_client.daemon_log_file)

    try:
        manager = get_manager()
        runner = manager.create_daemon_runner()
        manager.set_runner(runner)
    except Exception:
        LOGGER.exception('daemon runner failed to start')
        raise

    signals = (signal.SIGTERM, signal.SIGINT)
    for s in signals:  # pylint: disable=invalid-name
        runner.loop.add_signal_handler(s, lambda s=s: asyncio.create_task(shutdown_runner(runner)))

    try:
        LOGGER.info('Starting a daemon runner')
        runner.start()
    except SystemError as exception:
        LOGGER.info('Received a SystemError: %s', exception)
        runner.close()

    LOGGER.info('Daemon runner started')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Functions to dynamically generate reversed polish notation expressions."""
import collections
import operator as operators
import random

OPERATORS = {
    '+': operators.add,
    '*': operators.mul,
    '^': operators.pow,
}


def generate(min_operator_count=3, max_operator_count=5, min_operand_value=-5, max_operand_value=5):
    """
    Generate a random valid expression in Reverse Polish Notation. There are a few limitations:

        * Only integers are supported
        * Only the addition, multiplication and power operators (+, * and ^, respectively) are supported
        * Every intermediate result should be an integer, so no raising a number to a negative power
        * Operands for power operator are limited to the range [1, 3]
        * Expression has only a single sequence of numbers followed by single continuous sequence of operators

    :param min_operator_count: the minimum number of operators the expression should have
    :param max_operator_count: the maximum number of operators the expression should have
    :param min_operand_value: the minimum value of all the operands
    :param max_operand_value: the minimum value of all the operands
    :return: a string representing an expression in Reverse Polish Notation
    """
    operator_count = random.randint(min_operator_count, max_operator_count)
    operand_range = list(range(min_operand_value, max_operand_value + 1))
    operand_range.remove(0)
    operand_range_pow = [1, 3]

    # We start with the initial operand
    symbols = [str(random.choice(operand_range))]

    for _ in range(operator_count):
        operator = random.choice(OPERATORS.keys())

        if OPERATORS[operator] is operators.pow:
            operand = random.choice(operand_range_pow)
        else:
            operand = random.choice(operand_range)

        symbols.append(operator)
        symbols.insert(0, str(operand))

    return ' '.join(symbols)


def validate(expression):
    """
    Validate an expression in Reverse Polish Notation. In addition to normal rules, the following restrictions apply:

        * Only integers are supported
        * Only the addition, multiplication and power operators (+, * and ^, respectively) are supported
        * Every intermediate result should be an integer, so no raising a number to a negative power
        * Expression has only a single sequence of numbers followed by single continuous sequence of operators

    :param expression: the expression in Reverse Polish Notation
    :return: tuple(Bool, list) indicating whether expression is valid and if not a list of error messages
    """
    # pylint: disable=too-many-return-statements
    try:
        symbols = expression.split()
    except ValueError as exception:
        return False, f'failed to split the expression into symbols: {exception}'

    while len(symbols) > 1:
        try:
            operand = symbols.pop(0)
            operator = symbols.pop()
        except IndexError:
            return False, 'expression contains incongruent number of symbols'

        try:
            operand = int(operand)
        except ValueError:
            return False, f'the operand {operand} is not a valid integer'

        if operator not in OPERATORS.keys():
            return False, f'the operator {operator} is not supported'

        if OPERATORS[operator] is operators.pow and operand < 0:
            return False, f'a negative operand {operand} was found for the ^ operator, which is not allowed'

    # At this point the symbols list should only contain the initial operand
    try:
        operand = int(symbols.pop())
    except ValueError:
        return False, f'the operand {operand} is not a valid integer'

    if symbols:
        return False, 'incorrect number of symbols found, should contain N operands followed by (N - 1) operators'

    return True, None


def evaluate(expression, modulo=None):
    """
    Evaluate an expression in Reverse Polish Notation. There are a few limitations:

        * Only integers are supported
        * Only the addition, multiplication and power operators (+, * and ^, respectively) are supported

    :param expression: the expression in Reverse Polish Notation
    :return: the result of the evaluted expression
    """
    stack = collections.deque()
    result = 0

    for part in expression.split():
        if part not in OPERATORS.keys():
            stack.appendleft(part)
        else:
            operator = OPERATORS[part]

            left_hand = int(stack.popleft())
            right_hand = int(stack.popleft())
            result = operator(left_hand, right_hand)
            stack.appendleft(str(result))

    if modulo is not None:
        return int(result % modulo)

    return int(result)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for synchronizing the django and sqlalchemy schema."""
import alembic

from aiida.storage.psql_dos.migrations.utils import ReflectMigrations


def synchronize_schemas(alembic_op: alembic.op) -> None:
    """This function is used by the final migration step, of django/sqlalchemy branches, to synchronize their schemas.

    1. Remove and recreate all (non-unique) indexes, with standard names and postgresql ops.
    2. Remove and recreate all unique constraints, with standard names.
    3. Remove and recreate all foreign key constraints, with standard names and other rules.

    Schema naming conventions are defined ``aiida/storage/sqlalchemy/models/base.py::naming_convention``.

    Note we assume here that (a) all primary keys are already correct, and (b) there are no check constraints.
    """
    reflect = ReflectMigrations(alembic_op)

    # drop all current non-unique indexes, then add the new ones
    for tbl_name in (
        'db_dbauthinfo', 'db_dbcomment', 'db_dbcomputer', 'db_dbgroup', 'db_dbgroup_dbnodes', 'db_dblink', 'db_dblog',
        'db_dbnode', 'db_dbsetting', 'db_dbuser'
    ):
        reflect.drop_all_indexes(tbl_name)
    for name, tbl_name, column, psql_op in (
        ('ix_db_dbauthinfo_aiidauser_id', 'db_dbauthinfo', 'aiidauser_id', None),
        ('ix_db_dbauthinfo_dbcomputer_id', 'db_dbauthinfo', 'dbcomputer_id', None),
        ('ix_db_dbcomment_dbnode_id', 'db_dbcomment', 'dbnode_id', None),
        ('ix_db_dbcomment_user_id', 'db_dbcomment', 'user_id', None),
        ('ix_pat_db_dbcomputer_label', 'db_dbcomputer', 'label', 'varchar_pattern_ops'),
        ('ix_db_dbgroup_label', 'db_dbgroup', 'label', None),
        ('ix_pat_db_dbgroup_label', 'db_dbgroup', 'label', 'varchar_pattern_ops'),
        ('ix_db_dbgroup_type_string', 'db_dbgroup', 'type_string', None),
        ('ix_pat_db_dbgroup_type_string', 'db_dbgroup', 'type_string', 'varchar_pattern_ops'),
        ('ix_db_dbgroup_user_id', 'db_dbgroup', 'user_id', None),
        ('ix_db_dbgroup_dbnodes_dbgroup_id', 'db_dbgroup_dbnodes', 'dbgroup_id', None),
        ('ix_db_dbgroup_dbnodes_dbnode_id', 'db_dbgroup_dbnodes', 'dbnode_id', None),
        ('ix_db_dblink_input_id', 'db_dblink', 'input_id', None),
        ('ix_db_dblink_label', 'db_dblink', 'label', None),
        ('ix_pat_db_dblink_label', 'db_dblink', 'label', 'varchar_pattern_ops'),
        ('ix_db_dblink_output_id', 'db_dblink', 'output_id', None),
        ('ix_db_dblink_type', 'db_dblink', 'type', None),
        ('ix_pat_db_dblink_type', 'db_dblink', 'type', 'varchar_pattern_ops'),
        ('ix_db_dblog_dbnode_id', 'db_dblog', 'dbnode_id', None),
        ('ix_db_dblog_levelname', 'db_dblog', 'levelname', None),
        ('ix_pat_db_dblog_levelname', 'db_dblog', 'levelname', 'varchar_pattern_ops'),
        ('ix_db_dblog_loggername', 'db_dblog', 'loggername', None),
        ('ix_pat_db_dblog_loggername', 'db_dblog', 'loggername', 'varchar_pattern_ops'),
        ('ix_db_dbnode_ctime', 'db_dbnode', 'ctime', None),
        ('ix_db_dbnode_dbcomputer_id', 'db_dbnode', 'dbcomputer_id', None),
        ('ix_db_dbnode_label', 'db_dbnode', 'label', None),
        ('ix_pat_db_dbnode_label', 'db_dbnode', 'label', 'varchar_pattern_ops'),
        ('ix_db_dbnode_mtime', 'db_dbnode', 'mtime', None),
        ('ix_db_dbnode_process_type', 'db_dbnode', 'process_type', None),
        ('ix_pat_db_dbnode_process_type', 'db_dbnode', 'process_type', 'varchar_pattern_ops'),
        ('ix_db_dbnode_node_type', 'db_dbnode', 'node_type', None),
        ('ix_pat_db_dbnode_node_type', 'db_dbnode', 'node_type', 'varchar_pattern_ops'),
        ('ix_db_dbnode_user_id', 'db_dbnode', 'user_id', None),
        ('ix_pat_db_dbsetting_key', 'db_dbsetting', 'key', 'varchar_pattern_ops'),
        ('ix_pat_db_dbuser_email', 'db_dbuser', 'email', 'varchar_pattern_ops'),
    ):
        kwargs = {'unique': False}
        if psql_op is not None:
            kwargs['postgresql_ops'] = {column: psql_op}
        alembic_op.create_index(name, tbl_name, [column], **kwargs)

    # drop all current unique constraints, then add the new ones
    for tbl_name in (
        'db_dbauthinfo', 'db_dbcomment', 'db_dbcomputer', 'db_dbgroup', 'db_dbgroup_dbnodes', 'db_dblink', 'db_dblog',
        'db_dbnode', 'db_dbsetting', 'db_dbuser'
    ):
        reflect.drop_all_unique_constraints(tbl_name)
    reflect.reset_cache()
    for name, tbl_name, columns in (
        ('uq_db_dbauthinfo_aiidauser_id_dbcomputer_id', 'db_dbauthinfo', ('aiidauser_id', 'dbcomputer_id')),
        ('uq_db_dbcomment_uuid', 'db_dbcomment', ('uuid',)),
        ('uq_db_dbcomputer_label', 'db_dbcomputer', ('label',)),
        ('uq_db_dbcomputer_uuid', 'db_dbcomputer', ('uuid',)),
        ('uq_db_dbgroup_label_type_string', 'db_dbgroup', ('label', 'type_string')),
        ('uq_db_dbgroup_uuid', 'db_dbgroup', ('uuid',)),
        ('uq_db_dbgroup_dbnodes_dbgroup_id_dbnode_id', 'db_dbgroup_dbnodes', ('dbgroup_id', 'dbnode_id')),
        ('uq_db_dblog_uuid', 'db_dblog', ('uuid',)),
        ('uq_db_dbnode_uuid', 'db_dbnode', ('uuid',)),
        ('uq_db_dbsetting_key', 'db_dbsetting', ('key',)),
        ('uq_db_dbuser_email', 'db_dbuser', ('email',)),
    ):
        reflect.drop_indexes(tbl_name, columns, unique=True)  # drop any remaining indexes
        alembic_op.create_unique_constraint(name, tbl_name, columns)

    # drop all current foreign key constraints, then add the new ones
    for tbl_name in (
        'db_dbauthinfo', 'db_dbcomment', 'db_dbcomputer', 'db_dbgroup', 'db_dbgroup_dbnodes', 'db_dblink', 'db_dblog',
        'db_dbnode', 'db_dbsetting', 'db_dbuser'
    ):
        reflect.drop_all_foreign_keys(tbl_name)

    alembic_op.create_foreign_key(
        'fk_db_dbauthinfo_aiidauser_id_db_dbuser',
        'db_dbauthinfo',
        'db_dbuser',
        ['aiidauser_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbauthinfo_dbcomputer_id_db_dbcomputer',
        'db_dbauthinfo',
        'db_dbcomputer',
        ['dbcomputer_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbcomment_dbnode_id_db_dbnode',
        'db_dbcomment',
        'db_dbnode',
        ['dbnode_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbcomment_user_id_db_dbuser',
        'db_dbcomment',
        'db_dbuser',
        ['user_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'db_dbgroup_user_id_db_dbuser',
        'db_dbgroup',
        'db_dbuser',
        ['user_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbgroup_dbnodes_dbgroup_id_db_dbgroup',
        'db_dbgroup_dbnodes',
        'db_dbgroup',
        ['dbgroup_id'],
        ['id'],
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbgroup_dbnodes_dbnode_id_db_dbnode',
        'db_dbgroup_dbnodes',
        'db_dbnode',
        ['dbnode_id'],
        ['id'],
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dblink_input_id_db_dbnode',
        'db_dblink',
        'db_dbnode',
        ['input_id'],
        ['id'],
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dblink_output_id_db_dbnode',
        'db_dblink',
        'db_dbnode',
        ['output_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dblog_dbnode_id_db_dbnode',
        'db_dblog',
        'db_dbnode',
        ['dbnode_id'],
        ['id'],
        ondelete='CASCADE',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbnode_dbcomputer_id_db_dbcomputer',
        'db_dbnode',
        'db_dbcomputer',
        ['dbcomputer_id'],
        ['id'],
        ondelete='RESTRICT',
        deferrable=True,
        initially='DEFERRED',
    )
    alembic_op.create_foreign_key(
        'fk_db_dbnode_user_id_db_dbuser',
        'db_dbnode',
        'db_dbuser',
        ['user_id'],
        ['id'],
        ondelete='RESTRICT',
        deferrable=True,
        initially='DEFERRED',
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi database` commands."""
# pylint: disable=unused-argument

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options
from aiida.cmdline.utils import decorators


@verdi.group('database', hidden=True)
def verdi_database():
    """Inspect and manage the database.

    .. deprecated:: v2.0.0
    """


@verdi_database.command('version')
@decorators.deprecated_command(
    'This command has been deprecated and no longer has any effect. It will be removed soon from the CLI (in v2.1).\n'
    'The same information is now available through `verdi storage version`.\n'
)
def database_version():
    """Show the version of the database.

    The database version is defined by the tuple of the schema generation and schema revision.

    .. deprecated:: v2.0.0
    """


@verdi_database.command('migrate')
@options.FORCE()
@click.pass_context
@decorators.deprecated_command(
    'This command has been deprecated and will be removed soon (in v3.0). '
    'Please call `verdi storage migrate` instead.\n'
)
def database_migrate(ctx, force):
    """Migrate the database to the latest schema version.

    .. deprecated:: v2.0.0
    """
    from aiida.cmdline.commands.cmd_storage import storage_migrate
    ctx.forward(storage_migrate)


@verdi_database.group('integrity')
def verdi_database_integrity():
    """Check the integrity of the database and fix potential issues.

    .. deprecated:: v2.0.0
    """


@verdi_database_integrity.command('detect-duplicate-uuid')
@click.option(
    '-t',
    '--table',
    default='db_dbnode',
    type=click.Choice(('db_dbcomment', 'db_dbcomputer', 'db_dbgroup', 'db_dbnode')),
    help='The database table to operate on.'
)
@click.option(
    '-a', '--apply-patch', is_flag=True, help='Actually apply the proposed changes instead of performing a dry run.'
)
@decorators.deprecated_command(
    'This command has been deprecated and no longer has any effect. It will be removed soon from the CLI (in v2.1).\n'
    'For remaining available integrity checks, use `verdi storage integrity` instead.\n'
)
def detect_duplicate_uuid(table, apply_patch):
    """Detect and fix entities with duplicate UUIDs.

    Before aiida-core v1.0.0, there was no uniqueness constraint on the UUID column of the node table in the database
    and a few other tables as well. This made it possible to store multiple entities with identical UUIDs in the same
    table without the database complaining. This bug was fixed in aiida-core=1.0.0 by putting an explicit uniqueness
    constraint on UUIDs on the database level. However, this would leave databases created before this patch with
    duplicate UUIDs in an inconsistent state. This command will run an analysis to detect duplicate UUIDs in a given
    table and solve it by generating new UUIDs. Note that it will not delete or merge any rows.


    .. deprecated:: v2.0.0
    """


@verdi_database_integrity.command('detect-invalid-links')
@decorators.with_dbenv()
@decorators.deprecated_command(
    'This command has been deprecated and no longer has any effect. It will be removed soon from the CLI (in v2.1).\n'
    'For remaining available integrity checks, use `verdi storage integrity` instead.\n'
)
def detect_invalid_links():
    """Scan the database for invalid links.

    .. deprecated:: v2.0.0
    """


@verdi_database_integrity.command('detect-invalid-nodes')
@decorators.with_dbenv()
@decorators.deprecated_command(
    'This command has been deprecated and no longer has any effect. It will be removed soon from the CLI (in v2.1).\n'
    'For remaining available integrity checks, use `verdi storage integrity` instead.\n'
)
def detect_invalid_nodes():
    """Scan the database for invalid nodes.

    .. deprecated:: v2.0.0
    """


@verdi_database.command('summary')
@decorators.deprecated_command(
    'This command has been deprecated and no longer has any effect. It will be removed soon from the CLI (in v2.1).\n'
    'Please call `verdi storage info` instead.\n'
)
def database_summary():
    """Summarise the entities in the database.

    .. deprecated:: v2.0.0
    """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub class `Data` to be used as a base class for data structures."""
from aiida.common import exceptions
from aiida.common.lang import override
from aiida.common.links import LinkType

from ..node import Node

__all__ = ('Data',)


class Data(Node):
    """
    The base class for all Data nodes.

    AiiDA Data classes are subclasses of Node and must support multiple inheritance.

    Architecture note:
    Calculation plugins are responsible for converting raw output data from simulation codes to Data nodes.
    Nodes are responsible for validating their content (see _validate method).
    """
    _source_attributes = ['db_name', 'db_uri', 'uri', 'id', 'version', 'extras', 'source_md5', 'description', 'license']

    # Replace this with a dictionary in each subclass that, given a file
    # extension, returns the corresponding fileformat string.
    #
    # This is used in the self.export() method.
    # By default, if not found here,
    # The fileformat string is assumed to match the extension.
    # Example: {'dat': 'dat_multicolumn'}
    _export_format_replacements = {}

    # Data nodes are storable
    _storable = True
    _unstorable_message = 'storing for this node has been disabled'

    def __init__(self, *args, source=None, **kwargs):
        """Construct a new instance, setting the ``source`` attribute if provided as a keyword argument."""
        super().__init__(*args, **kwargs)
        if source is not None:
            self.source = source

    def __copy__(self):
        """Copying a Data node is not supported, use copy.deepcopy or call Data.clone()."""
        raise exceptions.InvalidOperation('copying a Data node is not supported, use copy.deepcopy')

    def __deepcopy__(self, memo):
        """
        Create a clone of the Data node by piping through to the clone method and return the result.

        :returns: an unstored clone of this Data node
        """
        return self.clone()

    def clone(self):
        """Create a clone of the Data node.

        :returns: an unstored clone of this Data node
        """
        import copy

        backend_clone = self.backend_entity.clone()
        clone = self.__class__.from_backend_entity(backend_clone)
        clone.base.attributes.reset(copy.deepcopy(self.base.attributes.all))
        clone.base.repository._clone(self.base.repository)  # pylint: disable=protected-access

        return clone

    @property
    def source(self):
        """
        Gets the dictionary describing the source of Data object. Possible fields:

        * **db_name**: name of the source database.
        * **db_uri**: URI of the source database.
        * **uri**: URI of the object's source. Should be a permanent link.
        * **id**: object's source identifier in the source database.
        * **version**: version of the object's source.
        * **extras**: a dictionary with other fields for source description.
        * **source_md5**: MD5 checksum of object's source.
        * **description**: human-readable free form description of the object's source.
        * **license**: a string with a type of license.

        .. note:: some limitations for setting the data source exist, see ``_validate`` method.

        :return: dictionary describing the source of Data object.
        """
        return self.base.attributes.get('source', None)

    @source.setter
    def source(self, source):
        """
        Sets the dictionary describing the source of Data object.

        :raise KeyError: if dictionary contains unknown field.
        :raise ValueError: if supplied source description is not a dictionary.
        """
        if not isinstance(source, dict):
            raise ValueError('Source must be supplied as a dictionary')
        unknown_attrs = tuple(set(source.keys()) - set(self._source_attributes))
        if unknown_attrs:
            raise KeyError(f"Unknown source parameters: {', '.join(unknown_attrs)}")

        self.base.attributes.set('source', source)

    def set_source(self, source):
        """
        Sets the dictionary describing the source of Data object.
        """
        self.source = source

    @property
    def creator(self):
        """Return the creator of this node or None if it does not exist.

        :return: the creating node or None
        """
        inputs = self.base.links.get_incoming(link_type=LinkType.CREATE)
        if inputs:
            return inputs.first().node

        return None

    @override
    def _exportcontent(self, fileformat, main_file_name='', **kwargs):
        """
        Converts a Data node to one (or multiple) files.

        Note: Export plugins should return utf8-encoded **bytes**, which can be
        directly dumped to file.

        :param fileformat: the extension, uniquely specifying the file format.
        :type fileformat: str
        :param main_file_name: (empty by default) Can be used by plugin to
            infer sensible names for additional files, if necessary.  E.g. if the
            main file is '../myplot.gnu', the plugin may decide to store the dat
            file under '../myplot_data.dat'.
        :type main_file_name: str
        :param kwargs: other parameters are passed down to the plugin
        :returns: a tuple of length 2. The first element is the content of the
            otuput file. The second is a dictionary (possibly empty) in the format
            {filename: filecontent} for any additional file that should be produced.
        :rtype: (bytes, dict)
        """
        exporters = self._get_exporters()

        try:
            func = exporters[fileformat]
        except KeyError:
            if exporters.keys():
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'Currently implemented are: {}.'.format(
                        fileformat, self.__class__.__name__, ','.join(exporters.keys())
                    )
                )
            else:
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'No formats are implemented yet.'.format(fileformat, self.__class__.__name__)
                )

        string, dictionary = func(main_file_name=main_file_name, **kwargs)
        assert isinstance(string, bytes), 'export function `{}` did not return the content as a byte string.'

        return string, dictionary

    @override
    def export(self, path, fileformat=None, overwrite=False, **kwargs):
        """
        Save a Data object to a file.

        :param fname: string with file name. Can be an absolute or relative path.
        :param fileformat: kind of format to use for the export. If not present,
            it will try to use the extension of the file name.
        :param overwrite: if set to True, overwrites file found at path. Default=False
        :param kwargs: additional parameters to be passed to the
            _exportcontent method
        :return: the list of files created
        """
        import os

        if not path:
            raise ValueError('Path not recognized')

        if os.path.exists(path) and not overwrite:
            raise OSError(f'A file was already found at {path}')

        if fileformat is None:
            extension = os.path.splitext(path)[1]
            if extension.startswith(os.path.extsep):
                extension = extension[len(os.path.extsep):]
            if not extension:
                raise ValueError('Cannot recognized the fileformat from the extension')

            # Replace the fileformat using the replacements specified in the
            # _export_format_replacements dictionary. If not found there,
            # by default assume the fileformat string is identical to the extension
            fileformat = self._export_format_replacements.get(extension, extension)

        retlist = []

        filetext, extra_files = self._exportcontent(fileformat, main_file_name=path, **kwargs)

        if not overwrite:
            for fname in extra_files:
                if os.path.exists(fname):
                    raise OSError(f'The file {fname} already exists, stopping.')

            if os.path.exists(path):
                raise OSError(f'The file {path} already exists, stopping.')

        for additional_fname, additional_fcontent in extra_files.items():
            retlist.append(additional_fname)
            with open(additional_fname, 'wb', encoding=None) as fhandle:
                fhandle.write(additional_fcontent)  # This is up to each specific plugin
        retlist.append(path)
        with open(path, 'wb', encoding=None) as fhandle:
            fhandle.write(filetext)

        return retlist

    def _get_exporters(self):
        """
        Get all implemented export formats.
        The convention is to find all _prepare_... methods.
        Returns a dictionary of method_name: method_function
        """
        # NOTE: To add support for a new format, write a new function called as
        # _prepare_"" with the name of the new format
        exporter_prefix = '_prepare_'
        valid_format_names = self.get_export_formats()
        valid_formats = {k: getattr(self, exporter_prefix + k) for k in valid_format_names}
        return valid_formats

    @classmethod
    def get_export_formats(cls):
        """
        Get the list of valid export format strings

        :return: a list of valid formats
        """
        exporter_prefix = '_prepare_'
        method_names = dir(cls)  # get list of class methods names
        valid_format_names = [
            i[len(exporter_prefix):] for i in method_names if i.startswith(exporter_prefix)
        ]  # filter them
        return sorted(valid_format_names)

    def importstring(self, inputstring, fileformat, **kwargs):
        """
        Converts a Data object to other text format.

        :param fileformat: a string (the extension) to describe the file format.
        :returns: a string with the structure description.
        """
        importers = self._get_importers()

        try:
            func = importers[fileformat]
        except KeyError:
            if importers.keys():
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'Currently implemented are: {}.'.format(
                        fileformat, self.__class__.__name__, ','.join(importers.keys())
                    )
                )
            else:
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'No formats are implemented yet.'.format(fileformat, self.__class__.__name__)
                )

        # func is bound to self by getattr in _get_importers()
        func(inputstring, **kwargs)

    def importfile(self, fname, fileformat=None):
        """
        Populate a Data object from a file.

        :param fname: string with file name. Can be an absolute or relative path.
        :param fileformat: kind of format to use for the export. If not present,
            it will try to use the extension of the file name.
        """
        if fileformat is None:
            fileformat = fname.split('.')[-1]
        with open(fname, 'r', encoding='utf8') as fhandle:  # reads in cwd, if fname is not absolute
            self.importstring(fhandle.read(), fileformat)

    def _get_importers(self):
        """
        Get all implemented import formats.
        The convention is to find all _parse_... methods.
        Returns a list of strings.
        """
        # NOTE: To add support for a new format, write a new function called as
        # _parse_"" with the name of the new format
        importer_prefix = '_parse_'
        method_names = dir(self)  # get list of class methods names
        valid_format_names = [i[len(importer_prefix):] for i in method_names if i.startswith(importer_prefix)]
        valid_formats = {k: getattr(self, importer_prefix + k) for k in valid_format_names}
        return valid_formats

    def convert(self, object_format=None, *args):
        """
        Convert the AiiDA StructureData into another python object

        :param object_format: Specify the output format
        """
        # pylint: disable=keyword-arg-before-vararg

        if object_format is None:
            raise ValueError('object_format must be provided')

        if not isinstance(object_format, str):
            raise ValueError('object_format should be a string')

        converters = self._get_converters()

        try:
            func = converters[object_format]
        except KeyError:
            if converters.keys():
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'Currently implemented are: {}.'.format(
                        object_format, self.__class__.__name__, ','.join(converters.keys())
                    )
                )
            else:
                raise ValueError(
                    'The format {} is not implemented for {}. '
                    'No formats are implemented yet.'.format(object_format, self.__class__.__name__)
                )

        return func(*args)

    def _get_converters(self):
        """
        Get all implemented converter formats.
        The convention is to find all _get_object_... methods.
        Returns a list of strings.
        """
        # NOTE: To add support for a new format, write a new function called as
        # _prepare_"" with the name of the new format
        exporter_prefix = '_get_object_'
        method_names = dir(self)  # get list of class methods names
        valid_format_names = [i[len(exporter_prefix):] for i in method_names if i.startswith(exporter_prefix)]
        valid_formats = {k: getattr(self, exporter_prefix + k) for k in valid_format_names}
        return valid_formats



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""A utility module with a factory of standard QueryBuilder instances for Calculation nodes."""
from aiida.cmdline.utils.query.mapping import CalculationProjectionMapper
from aiida.common.lang import classproperty


class CalculationQueryBuilder:
    """Utility class to construct a QueryBuilder instance for Calculation nodes and project the query set."""

    # This tuple serves to mark compound projections that cannot explicitly be projected in the QueryBuilder, but will
    # have to be manually projected from composing its individual projection constituents
    _compound_projections = ('state',)
    _default_projections = ('pk', 'ctime', 'process_label', 'state', 'process_status')
    _valid_projections = (
        'pk', 'uuid', 'ctime', 'mtime', 'state', 'process_state', 'process_status', 'exit_status', 'sealed',
        'process_label', 'label', 'description', 'node_type', 'paused', 'process_type', 'job_state', 'scheduler_state',
        'exception'
    )

    def __init__(self, mapper=None):
        if mapper is None:
            self._mapper = CalculationProjectionMapper(self._valid_projections)
        else:
            self._mapper = mapper

    @property
    def mapper(self):
        return self._mapper

    @classproperty
    def default_projections(self):
        return self._default_projections

    @classproperty
    def valid_projections(self):
        return self._valid_projections

    def get_filters(
        self,
        all_entries=False,
        process_state=None,
        process_label=None,
        paused=False,
        exit_status=None,
        failed=False,
        node_types=None
    ):
        """
        Return a set of QueryBuilder filters based on typical command line options.

        :param node_types: a tuple of node classes to filter for (must be sub classes of Calculation)
        :param all_entries: boolean to negate filtering for process state
        :param process_state: filter for this process state attribute
        :param process_label: filter for this process label attribute
        :param paused: boolean, if True, filter for processes that are paused
        :param exit_status: filter for this exit status
        :param failed: boolean to filter only failed processes
        :return: dictionary of filters suitable for a QueryBuilder.append() call
        """
        # pylint: disable=too-many-arguments
        from aiida.engine import ProcessState

        exit_status_attribute = self.mapper.get_attribute('exit_status')
        process_label_attribute = self.mapper.get_attribute('process_label')
        process_state_attribute = self.mapper.get_attribute('process_state')
        paused_attribute = self.mapper.get_attribute('paused')

        filters = {}

        if node_types is not None:
            filters['or'] = []
            for node_class in node_types:
                filters['or'].append({'type': node_class.class_node_type})

        if process_state and not all_entries:
            filters[process_state_attribute] = {'in': process_state}

        if process_label is not None:
            if '%' in process_label or '_' in process_label:
                filters[process_label_attribute] = {'like': process_label}
            else:
                filters[process_label_attribute] = process_label

        if paused:
            filters[paused_attribute] = True

        if failed:
            filters[process_state_attribute] = {'==': ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {'>': 0}

        if exit_status is not None:
            filters[process_state_attribute] = {'==': ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {'==': exit_status}

        return filters

    def get_query_set(self, relationships=None, filters=None, order_by=None, past_days=None, limit=None):
        """
        Return the query set of calculations for the given filters and query parameters

        :param relationships: a mapping of relationships to join on, e.g. {'with_node': Group} to join on a Group. The
            keys in this dictionary should be the keyword used in the `append` method of the `QueryBuilder` to join the
            entity on that is defined as the value.
        :param filters: rules to filter query results with
        :param order_by: order the query set by this criterion
        :param past_days: only include entries from the last past days
        :param limit: limit the query set to this number of entries
        :return: the query set, a list of dictionaries
        """
        import datetime

        from aiida import orm
        from aiida.common import timezone

        # Define the list of projections for the QueryBuilder, which are all valid minus the compound projections
        projected_attributes = [
            self.mapper.get_attribute(projection)
            for projection in self._valid_projections
            if projection not in self._compound_projections
        ]

        if filters is None:
            filters = {}

        if past_days is not None:
            filters['ctime'] = {'>': timezone.now() - datetime.timedelta(days=past_days)}

        builder = orm.QueryBuilder()
        builder.append(cls=orm.ProcessNode, filters=filters, project=projected_attributes, tag='process')

        if relationships is not None:
            for tag, entity in relationships.items():
                builder.append(cls=type(entity), filters={'id': entity.pk}, **{tag: 'process'})

        if order_by is not None:
            builder.order_by({'process': order_by})
        else:
            builder.order_by({'process': {'ctime': 'asc'}})

        if limit is not None:
            builder.limit(limit)

        return builder.iterdict()

    def get_projected(self, query_set, projections):
        """
        Project the query set for the given set of projections
        """
        header = [self.mapper.get_label(projection) for projection in projections]
        result = [header]

        for query_result in query_set:
            result_row = [self.mapper.format(projection, query_result['process']) for projection in projections]
            result.append(result_row)

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for all common top level AiiDA entity classes and methods"""
import abc
from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Generic, List, Optional, Type, TypeVar, cast

from plumpy.base.utils import call_with_super_check, super_check

from aiida.common.lang import classproperty, type_check
from aiida.common.warnings import warn_deprecation
from aiida.manage import get_manager

if TYPE_CHECKING:
    from aiida.orm.implementation import BackendEntity, StorageBackend
    from aiida.orm.querybuilder import FilterType, OrderByType, QueryBuilder

__all__ = ('Entity', 'Collection', 'EntityTypes')

CollectionType = TypeVar('CollectionType', bound='Collection')
EntityType = TypeVar('EntityType', bound='Entity')
BackendEntityType = TypeVar('BackendEntityType', bound='BackendEntity')


class EntityTypes(Enum):
    """Enum for referring to ORM entities in a backend-agnostic manner."""
    AUTHINFO = 'authinfo'
    COMMENT = 'comment'
    COMPUTER = 'computer'
    GROUP = 'group'
    LOG = 'log'
    NODE = 'node'
    USER = 'user'
    LINK = 'link'
    GROUP_NODE = 'group_node'


class Collection(abc.ABC, Generic[EntityType]):
    """Container class that represents the collection of objects of a particular entity type."""

    @staticmethod
    @abc.abstractmethod
    def _entity_base_cls() -> Type[EntityType]:
        """The allowed entity class or subclasses thereof."""

    @classmethod
    @lru_cache(maxsize=100)
    def get_cached(cls, entity_class: Type[EntityType], backend: 'StorageBackend'):
        """Get the cached collection instance for the given entity class and backend.

        :param backend: the backend instance to get the collection for
        """
        from aiida.orm.implementation import StorageBackend
        type_check(backend, StorageBackend)
        return cls(entity_class, backend=backend)

    def __init__(self, entity_class: Type[EntityType], backend: Optional['StorageBackend'] = None) -> None:
        """ Construct a new entity collection.

        :param entity_class: the entity type e.g. User, Computer, etc
        :param backend: the backend instance to get the collection for, or use the default
        """
        from aiida.orm.implementation import StorageBackend
        type_check(backend, StorageBackend, allow_none=True)
        assert issubclass(entity_class, self._entity_base_cls())
        self._backend = backend or get_manager().get_profile_storage()
        self._entity_type = entity_class

    def __call__(self: CollectionType, backend: 'StorageBackend') -> CollectionType:
        """Get or create a cached collection using a new backend."""
        if backend is self._backend:
            return self
        return self.get_cached(self.entity_type, backend=backend)  # type: ignore

    @property
    def entity_type(self) -> Type[EntityType]:
        """The entity type for this instance."""
        return self._entity_type

    @property
    def backend(self) -> 'StorageBackend':
        """Return the backend."""
        return self._backend

    def query(
        self,
        filters: Optional['FilterType'] = None,
        order_by: Optional['OrderByType'] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None
    ) -> 'QueryBuilder':
        """Get a query builder for the objects of this collection.

        :param filters: the keyword value pair filters to match
        :param order_by: a list of (key, direction) pairs specifying the sort order
        :param limit: the maximum number of results to return
        :param offset: number of initial results to be skipped
        """
        from . import querybuilder

        filters = filters or {}
        order_by = {self.entity_type: order_by} if order_by else {}

        query = querybuilder.QueryBuilder(backend=self._backend, limit=limit, offset=offset)
        query.append(self.entity_type, project='*', filters=filters)
        query.order_by([order_by])
        return query

    def get(self, **filters: Any) -> EntityType:
        """Get a single collection entry that matches the filter criteria.

        :param filters: the filters identifying the object to get

        :return: the entry
        """
        res = self.query(filters=filters)
        return res.one()[0]

    def find(
        self,
        filters: Optional['FilterType'] = None,
        order_by: Optional['OrderByType'] = None,
        limit: Optional[int] = None
    ) -> List[EntityType]:
        """Find collection entries matching the filter criteria.

        :param filters: the keyword value pair filters to match
        :param order_by: a list of (key, direction) pairs specifying the sort order
        :param limit: the maximum number of results to return

        :return: a list of resulting matches
        """
        query = self.query(filters=filters, order_by=order_by, limit=limit)
        return cast(List[EntityType], query.all(flat=True))

    def all(self) -> List[EntityType]:
        """Get all entities in this collection.

        :return: A list of all entities
        """
        return cast(List[EntityType], self.query().all(flat=True))  # pylint: disable=no-member

    def count(self, filters: Optional['FilterType'] = None) -> int:
        """Count entities in this collection according to criteria.

        :param filters: the keyword value pair filters to match

        :return: The number of entities found using the supplied criteria
        """
        return self.query(filters=filters).count()


class Entity(abc.ABC, Generic[BackendEntityType]):
    """An AiiDA entity"""

    _CLS_COLLECTION = Collection

    @classproperty
    def objects(cls: EntityType) -> Collection[EntityType]:  # pylint: disable=no-self-argument,no-self-use
        """Get a collection for objects of this type, with the default backend.

        .. deprecated:: This will be removed in v3, use ``collection`` instead.

        :return: an object that can be used to access entities of this type
        """
        warn_deprecation('This property is deprecated, use `.collection` instead.', version=3, stacklevel=2)
        return cls.collection

    @classproperty
    def collection(cls: EntityType) -> Collection[EntityType]:  # pylint: disable=no-self-argument,no-self-use
        """Get a collection for objects of this type, with the default backend.

        :return: an object that can be used to access entities of this type
        """
        return cls._CLS_COLLECTION.get_cached(cls, get_manager().get_profile_storage())

    @classmethod
    def get(cls, **kwargs):
        """Get an entity of the collection matching the given filters.

        .. deprecated: Will be removed in v3, use `Entity.collection.get` instead.

        """
        warn_deprecation(
            f'This method is deprecated, use `{cls.__name__}.collection.get` instead.', version=3, stacklevel=2
        )
        return cls.collection.get(**kwargs)  # pylint: disable=no-member

    @classmethod
    def from_backend_entity(cls: Type[EntityType], backend_entity: BackendEntityType) -> EntityType:
        """Construct an entity from a backend entity instance

        :param backend_entity: the backend entity

        :return: an AiiDA entity instance
        """
        from .implementation.entities import BackendEntity

        type_check(backend_entity, BackendEntity)
        entity = cls.__new__(cls)
        entity._backend_entity = backend_entity
        call_with_super_check(entity.initialize)
        return entity

    def __init__(self, backend_entity: BackendEntityType) -> None:
        """
        :param backend_entity: the backend model supporting this entity
        """
        self._backend_entity = backend_entity
        call_with_super_check(self.initialize)

    @super_check
    def initialize(self) -> None:
        """Initialize instance attributes.

        This will be called after the constructor is called or an entity is created from an existing backend entity.
        """

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        """Return the id for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        .. deprecated: Will be removed in v3, use `pk` instead.

        :return: the entity's id
        """
        warn_deprecation('This method is deprecated, use `pk` instead.', version=3, stacklevel=2)
        return self._backend_entity.id

    @property
    def pk(self) -> int:
        """Return the primary key for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's principal key
        """
        return self._backend_entity.id

    def store(self: EntityType) -> EntityType:
        """Store the entity."""
        self._backend_entity.store()
        return self

    @property
    def is_stored(self) -> bool:
        """Return whether the entity is stored."""
        return self._backend_entity.is_stored

    @property
    def backend(self) -> 'StorageBackend':
        """Get the backend for this entity"""
        return self._backend_entity.backend

    @property
    def backend_entity(self) -> BackendEntityType:
        """Get the implementing class for this object"""
        return self._backend_entity



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Backend specific computer objects and methods"""
import abc
import logging
from typing import Any, Dict

from .entities import BackendCollection, BackendEntity

__all__ = ('BackendComputer', 'BackendComputerCollection')


class BackendComputer(BackendEntity):
    """Backend implementation for the `Computer` ORM class.

    A computer is a resource that can be used to run calculations:
    It has an associated transport_type, which points to a plugin for connecting to the resource and passing data,
    and a scheduler_type, which points to a plugin for scheduling calculations.
    """
    # pylint: disable=too-many-public-methods

    _logger = logging.getLogger(__name__)

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        """Return the UUID of the computer."""

    @property
    @abc.abstractmethod
    def label(self) -> str:
        """Return the (unique) label of the computer."""

    @abc.abstractmethod
    def set_label(self, val: str):
        """Set the (unique) label of the computer."""

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """Return the description of the computer."""

    @abc.abstractmethod
    def set_description(self, val: str):
        """Set the description of the computer."""

    @property
    @abc.abstractmethod
    def hostname(self) -> str:
        """Return the hostname of the computer (used to associate the connected device)."""

    @abc.abstractmethod
    def set_hostname(self, val: str) -> None:
        """
        Set the hostname of this computer
        :param val: The new hostname
        """

    @abc.abstractmethod
    def get_metadata(self) -> Dict[str, Any]:
        """Return the metadata for the computer."""

    @abc.abstractmethod
    def set_metadata(self, metadata: Dict[str, Any]) -> None:
        """Set the metadata for the computer."""

    @abc.abstractmethod
    def get_scheduler_type(self) -> str:
        """Return the scheduler plugin type."""

    @abc.abstractmethod
    def set_scheduler_type(self, scheduler_type: str) -> None:
        """Set the scheduler plugin type."""

    @abc.abstractmethod
    def get_transport_type(self) -> str:
        """Return the transport plugin type."""

    @abc.abstractmethod
    def set_transport_type(self, transport_type: str) -> None:
        """Set the transport plugin type."""

    @abc.abstractmethod
    def copy(self) -> 'BackendComputer':
        """Create an un-stored clone of an already stored `Computer`.

        :raises: ``InvalidOperation`` if the computer is not stored.
        """


class BackendComputerCollection(BackendCollection[BackendComputer]):
    """The collection of Computer entries."""

    ENTITY_CLASS = BackendComputer

    @abc.abstractmethod
    def delete(self, pk: int) -> None:
        """
        Delete an entry with the given pk

        :param pk: the pk of the entry to delete
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA specific implementation of plumpy Ports and PortNamespaces for the ProcessSpec."""
from collections.abc import Mapping
import re
from typing import Any, Callable, Dict, Optional, Sequence
import warnings

from plumpy import ports
from plumpy.ports import breadcrumbs_to_port

from aiida.common.links import validate_link_label
from aiida.orm import Data, Node

__all__ = (
    'PortNamespace', 'InputPort', 'OutputPort', 'CalcJobOutputPort', 'WithNonDb', 'WithSerialize',
    'PORT_NAMESPACE_SEPARATOR'
)

PORT_NAME_MAX_CONSECUTIVE_UNDERSCORES = 1  # pylint: disable=invalid-name
PORT_NAMESPACE_SEPARATOR = '__'  # The character sequence to represent a nested port namespace in a flat link label
OutputPort = ports.OutputPort  # pylint: disable=invalid-name


class WithNonDb:
    """
    A mixin that adds support to a port to flag that it should not be stored
    in the database using the non_db=True flag.

    The mixins have to go before the main port class in the superclass order
    to make sure the mixin has the chance to strip out the non_db keyword.
    """

    def __init__(self, *args, **kwargs) -> None:
        self._non_db_explicitly_set: bool = bool('non_db' in kwargs)
        non_db = kwargs.pop('non_db', False)
        super().__init__(*args, **kwargs)
        self._non_db: bool = non_db

    @property
    def non_db_explicitly_set(self) -> bool:
        """Return whether the a value for `non_db` was explicitly passed in the construction of the `Port`.

        :return: boolean, True if `non_db` was explicitly defined during construction, False otherwise
        """
        return self._non_db_explicitly_set

    @property
    def non_db(self) -> bool:
        """Return whether the value of this `Port` should be stored as a `Node` in the database.

        :return: boolean, True if it should be storable as a `Node`, False otherwise
        """
        return self._non_db

    @non_db.setter
    def non_db(self, non_db: bool) -> None:
        """Set whether the value of this `Port` should be stored as a `Node` in the database.
        """
        self._non_db_explicitly_set = True
        self._non_db = non_db


class WithSerialize:
    """
    A mixin that adds support for a serialization function which is automatically applied on inputs
    that are not AiiDA data types.
    """

    def __init__(self, *args, **kwargs) -> None:
        serializer = kwargs.pop('serializer', None)
        super().__init__(*args, **kwargs)
        self._serializer: Callable[[Any], 'Data'] = serializer

    def serialize(self, value: Any) -> 'Data':
        """Serialize the given value if it is not already a Data type and a serializer function is defined

        :param value: the value to be serialized
        :returns: a serialized version of the value or the unchanged value
        """
        if self._serializer is None or isinstance(value, Data):
            return value

        return self._serializer(value)


class InputPort(WithSerialize, WithNonDb, ports.InputPort):
    """
    Sub class of plumpy.InputPort which mixes in the WithSerialize and WithNonDb mixins to support automatic
    value serialization to database storable types and support non database storable input types as well.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Override the constructor to check the type of the default if set and warn if not immutable."""
        # pylint: disable=redefined-builtin,too-many-arguments
        if 'default' in kwargs:
            default = kwargs['default']
            # If the default is specified and it is a node instance, raise a warning. This is to try and prevent that
            # people set node instances as defaults which can cause various problems.
            if default is not ports.UNSPECIFIED and isinstance(default, Node):
                message = 'default of input port `{}` is a `Node` instance, which can lead to unexpected side effects.'\
                    ' It is advised to use a lambda instead, e.g.: `default=lambda: orm.Int(5)`.'.format(args[0])
                warnings.warn(UserWarning(message))  # pylint: disable=no-member

        super().__init__(*args, **kwargs)

    def get_description(self) -> Dict[str, str]:
        """
        Return a description of the InputPort, which will be a dictionary of its attributes

        :returns: a dictionary of the stringified InputPort attributes
        """
        description = super().get_description()
        description['non_db'] = f'{self.non_db}'

        return description


class CalcJobOutputPort(ports.OutputPort):
    """Sub class of plumpy.OutputPort which adds the `_pass_to_parser` attribute."""

    def __init__(self, *args, **kwargs) -> None:
        pass_to_parser = kwargs.pop('pass_to_parser', False)
        super().__init__(*args, **kwargs)
        self._pass_to_parser: bool = pass_to_parser

    @property
    def pass_to_parser(self) -> bool:
        return self._pass_to_parser


class PortNamespace(WithNonDb, ports.PortNamespace):
    """
    Sub class of plumpy.PortNamespace which implements the serialize method to support automatic recursive
    serialization of a given mapping onto the ports of the PortNamespace.
    """

    def __setitem__(self, key: str, port: ports.Port) -> None:
        """Ensure that a `Port` being added inherits the `non_db` attribute if not explicitly defined at construction.

        The reasoning is that if a `PortNamespace` has `non_db=True`, which is different from the default value, very
        often all leaves should be also `non_db=True`. To prevent a user from having to specify it manually everytime
        we overload the value here, unless it was specifically set during construction.

        Note that the `non_db` attribute is not present for all `Port` sub classes so we have to check for it first.
        """
        if not isinstance(port, ports.Port):
            raise TypeError('port needs to be an instance of Port')

        self.validate_port_name(key)

        if hasattr(port, 'non_db_explicitly_set') and not port.non_db_explicitly_set:  # type: ignore[attr-defined]
            port.non_db = self.non_db  # type: ignore[attr-defined]

        super().__setitem__(key, port)

    @staticmethod
    def validate_port_name(port_name: str) -> None:
        """Validate the given port name.

        Valid port names adhere to the following restrictions:

            * Is a valid link label (see below)
            * Does not contain two or more consecutive underscores

        Valid link labels adhere to the following restrictions:

            * Has to be a valid python identifier
            * Can only contain alphanumeric characters and underscores
            * Can not start or end with an underscore

        :param port_name: the proposed name of the port to be added
        :raise TypeError: if the port name is not a string type
        :raise ValueError: if the port name is invalid
        """
        try:
            validate_link_label(port_name)
        except ValueError as exception:
            raise ValueError(f'invalid port name `{port_name}`: {exception}')

        # Following regexes will match all groups of consecutive underscores where each group will be of the form
        # `('___', '_')`, where the first element is the matched group of consecutive underscores.
        consecutive_underscores = [match[0] for match in re.findall(r'((_)\2+)', port_name)]

        if any(len(entry) > PORT_NAME_MAX_CONSECUTIVE_UNDERSCORES for entry in consecutive_underscores):
            raise ValueError(f'invalid port name `{port_name}`: more than two consecutive underscores')

    def serialize(self, mapping: Optional[Dict[str, Any]], breadcrumbs: Sequence[str] = ()) -> Optional[Dict[str, Any]]:
        """Serialize the given mapping onto this `Portnamespace`.

        It will recursively call this function on any nested `PortNamespace` or the serialize function on any `Ports`.

        :param mapping: a mapping of values to be serialized
        :param breadcrumbs: a tuple with the namespaces of parent namespaces
        :returns: the serialized mapping
        """
        if mapping is None:
            return None

        breadcrumbs = (*breadcrumbs, self.name)

        if not isinstance(mapping, Mapping):
            port_name = breadcrumbs_to_port(breadcrumbs)
            raise TypeError(f'port namespace `{port_name}` received `{type(mapping)}` instead of a dictionary')

        result = {}

        for name, value in mapping.items():
            if name in self:

                port = self[name]
                if isinstance(port, PortNamespace):
                    result[name] = port.serialize(value, breadcrumbs)
                elif isinstance(port, InputPort):
                    result[name] = port.serialize(value)
                else:
                    raise AssertionError(f'port does not have a serialize method: {port}')
            else:
                result[name] = value

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Class and decorators to generate processes out of simple python functions."""
import collections
import functools
import inspect
import logging
import signal
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Sequence, Tuple, Type

from aiida.common.lang import override
from aiida.manage import get_manager
from aiida.orm import CalcFunctionNode, Data, ProcessNode, WorkFunctionNode
from aiida.orm.utils.mixins import FunctionCalculationMixin

from .process import Process

if TYPE_CHECKING:
    from .exit_code import ExitCode

__all__ = ('calcfunction', 'workfunction', 'FunctionProcess')

LOGGER = logging.getLogger(__name__)


def calcfunction(function: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator to turn a standard python function into a calcfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the calcfunction
    >>> @calcfunction
    >>> def sum(a, b):
    >>>    return a + b
    >>> # Run it with some input
    >>> r = sum(Int(4), Int(5))
    >>> print(r)
    9
    >>> r.base.links.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<CalcFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.base.links.get_incoming().get_node_by_label('result').base.links.get_incoming().all_nodes()
    [4, 5]

    :param function: The function to decorate.
    :type function: callable

    :return: The decorated function.
    :rtype: callable
    """
    return process_function(node_class=CalcFunctionNode)(function)


def workfunction(function: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator to turn a standard python function into a workfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the workfunction
    >>> @workfunction
    >>> def select(a, b):
    >>>    return a
    >>> # Run it with some input
    >>> r = select(Int(4), Int(5))
    >>> print(r)
    4
    >>> r.base.links.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<WorkFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.base.links.get_incoming().get_node_by_label('result').base.links.get_incoming().all_nodes()
    [4, 5]

    :param function: The function to decorate.
    :type function: callable

    :return: The decorated function.

    """
    return process_function(node_class=WorkFunctionNode)(function)


def process_function(node_class: Type['ProcessNode']) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    The base function decorator to create a FunctionProcess out of a normal python function.

    :param node_class: the ORM class to be used as the Node record for the FunctionProcess
    :type node_class: :class:`aiida.orm.ProcessNode`
    """

    def decorator(function: Callable[..., Any]) -> Callable[..., Any]:
        """
        Turn the decorated function into a FunctionProcess.

        :param callable function: the actual decorated function that the FunctionProcess represents
        :return callable: The decorated function.
        """
        process_class = FunctionProcess.build(function, node_class=node_class)

        def run_get_node(*args, **kwargs) -> Tuple[Optional[Dict[str, Any]], 'ProcessNode']:
            """
            Run the FunctionProcess with the supplied inputs in a local runner.

            :param args: input arguments to construct the FunctionProcess
            :param kwargs: input keyword arguments to construct the FunctionProcess
            :return: tuple of the outputs of the process and the process node

            """
            manager = get_manager()
            runner = manager.get_runner()
            inputs = process_class.create_inputs(*args, **kwargs)

            # Remove all the known inputs from the kwargs
            for port in process_class.spec().inputs:
                kwargs.pop(port, None)

            # If any kwargs remain, the spec should be dynamic, so we raise if it isn't
            if kwargs and not process_class.spec().inputs.dynamic:
                raise ValueError(f'{function.__name__} does not support these kwargs: {kwargs.keys()}')

            process = process_class(inputs=inputs, runner=runner)

            # Only add handlers for interrupt signal to kill the process if we are in a local and not a daemon runner.
            # Without this check, running process functions in a daemon worker would be killed if the daemon is shutdown
            current_runner = manager.get_runner()
            original_handler = None
            kill_signal = signal.SIGINT

            if not current_runner.is_daemon_runner:

                def kill_process(_num, _frame):
                    """Send the kill signal to the process in the current scope."""
                    LOGGER.critical('runner received interrupt, killing process %s', process.pid)
                    result = process.kill(msg='Process was killed because the runner received an interrupt')
                    return result

                # Store the current handler on the signal such that it can be restored after process has terminated
                original_handler = signal.getsignal(kill_signal)
                signal.signal(kill_signal, kill_process)

            try:
                result = process.execute()
            finally:
                # If the `original_handler` is set, that means the `kill_process` was bound, which needs to be reset
                if original_handler:
                    signal.signal(signal.SIGINT, original_handler)

            store_provenance = inputs.get('metadata', {}).get('store_provenance', True)
            if not store_provenance:
                process.node._storable = False  # pylint: disable=protected-access
                process.node._unstorable_message = 'cannot store node because it was run with `store_provenance=False`'  # pylint: disable=protected-access

            return result, process.node

        def run_get_pk(*args, **kwargs) -> Tuple[Optional[Dict[str, Any]], int]:
            """Recreate the `run_get_pk` utility launcher.

            :param args: input arguments to construct the FunctionProcess
            :param kwargs: input keyword arguments to construct the FunctionProcess
            :return: tuple of the outputs of the process and the process node pk

            """
            result, node = run_get_node(*args, **kwargs)
            return result, node.pk

        @functools.wraps(function)
        def decorated_function(*args, **kwargs):
            """This wrapper function is the actual function that is called."""
            result, _ = run_get_node(*args, **kwargs)
            return result

        decorated_function.run = decorated_function  # type: ignore[attr-defined]
        decorated_function.run_get_pk = run_get_pk  # type: ignore[attr-defined]
        decorated_function.run_get_node = run_get_node  # type: ignore[attr-defined]
        decorated_function.is_process_function = True  # type: ignore[attr-defined]
        decorated_function.node_class = node_class  # type: ignore[attr-defined]
        decorated_function.process_class = process_class  # type: ignore[attr-defined]
        decorated_function.recreate_from = process_class.recreate_from  # type: ignore[attr-defined]
        decorated_function.spec = process_class.spec  # type: ignore[attr-defined]

        return decorated_function

    return decorator


class FunctionProcess(Process):
    """Function process class used for turning functions into a Process"""

    _func_args: Sequence[str] = ()

    @staticmethod
    def _func(*_args, **_kwargs) -> dict:
        """
        This is used internally to store the actual function that is being
        wrapped and will be replaced by the build method.
        """
        return {}

    @staticmethod
    def build(func: Callable[..., Any], node_class: Type['ProcessNode']) -> Type['FunctionProcess']:
        """
        Build a Process from the given function.

        All function arguments will be assigned as process inputs. If keyword arguments are specified then
        these will also become inputs.

        :param func: The function to build a process from
        :param node_class: Provide a custom node class to be used, has to be constructable with no arguments. It has to
            be a sub class of `ProcessNode` and the mixin :class:`~aiida.orm.utils.mixins.FunctionCalculationMixin`.

        :return: A Process class that represents the function

        """
        if not issubclass(node_class, ProcessNode) or not issubclass(node_class, FunctionCalculationMixin):
            raise TypeError('the node_class should be a sub class of `ProcessNode` and `FunctionCalculationMixin`')

        args, varargs, keywords, defaults, _, _, _ = inspect.getfullargspec(func)
        nargs = len(args)
        ndefaults = len(defaults) if defaults else 0
        first_default_pos = nargs - ndefaults

        if varargs is not None:
            raise ValueError('variadic arguments are not supported')

        def _define(cls, spec):  # pylint: disable=unused-argument
            """Define the spec dynamically"""
            super().define(spec)

            for i, arg in enumerate(args):
                default = ()
                if defaults and i >= first_default_pos:
                    default = defaults[i - first_default_pos]

                # If the keyword was already specified, simply override the default
                if spec.has_input(arg):
                    spec.inputs[arg].default = default
                else:
                    # If the default is `None` make sure that the port also accepts a `NoneType`
                    # Note that we cannot use `None` because the validation will call `isinstance` which does not work
                    # when passing `None`, but it does work with `NoneType` which is returned by calling `type(None)`
                    if default is None:
                        valid_type = (Data, type(None))
                    else:
                        valid_type = (Data,)

                    spec.input(arg, valid_type=valid_type, default=default)

            # Set defaults for label and description based on function name and docstring, if not explicitly defined
            port_label = spec.inputs['metadata']['label']

            if not port_label.has_default():
                port_label.default = func.__name__

            # If the function support kwargs then allow dynamic inputs, otherwise disallow
            spec.inputs.dynamic = keywords is not None

            # Function processes must have a dynamic output namespace since we do not know beforehand what outputs
            # will be returned and the valid types for the value should be `Data` nodes as well as a dictionary because
            # the output namespace can be nested.
            spec.outputs.valid_type = (Data, dict)

        return type(
            func.__name__, (FunctionProcess,), {
                '__module__': func.__module__,
                '__name__': func.__name__,
                '_func': staticmethod(func),
                Process.define.__name__: classmethod(_define),
                '_func_args': args,
                '_node_class': node_class
            }
        )

    @classmethod
    def validate_inputs(cls, *args: Any, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """
        Validate the positional and keyword arguments passed in the function call.

        :raises TypeError: if more positional arguments are passed than the function defines
        """
        nargs = len(args)
        nparameters = len(cls._func_args)

        # If the spec is dynamic, i.e. the function signature includes `**kwargs` and the number of positional arguments
        # passed is larger than the number of explicitly defined parameters in the signature, the inputs are invalid and
        # we should raise. If we don't, some of the passed arguments, intended to be positional arguments, will be
        # misinterpreted as keyword arguments, but they won't have an explicit name to use for the link label, causing
        # the input link to be completely lost.
        if cls.spec().inputs.dynamic and nargs > nparameters:
            name = cls._func.__name__
            raise TypeError(f'{name}() takes {nparameters} positional arguments but {nargs} were given')

    @classmethod
    def create_inputs(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """Create the input args for the FunctionProcess."""
        cls.validate_inputs(*args, **kwargs)

        ins = {}
        if kwargs:
            ins.update(kwargs)
        if args:
            ins.update(cls.args_to_dict(*args))
        return ins

    @classmethod
    def args_to_dict(cls, *args: Any) -> Dict[str, Any]:
        """
        Create an input dictionary (of form label -> value) from supplied args.

        :param args: The values to use for the dictionary

        :return: A label -> value dictionary

        """
        return dict(list(zip(cls._func_args, args)))

    @classmethod
    def get_or_create_db_record(cls) -> 'ProcessNode':
        return cls._node_class()

    def __init__(self, *args, **kwargs) -> None:
        if kwargs.get('enable_persistence', False):
            raise RuntimeError('Cannot persist a function process')
        super().__init__(enable_persistence=False, *args, **kwargs)  # type: ignore

    @property
    def process_class(self) -> Callable[..., Any]:
        """
        Return the class that represents this Process, for the FunctionProcess this is the function itself.

        For a standard Process or sub class of Process, this is the class itself. However, for legacy reasons,
        the Process class is a wrapper around another class. This function returns that original class, i.e. the
        class that really represents what was being executed.

        :return: A Process class that represents the function

        """
        return self._func

    def execute(self) -> Optional[Dict[str, Any]]:
        """Execute the process."""
        result = super().execute()

        # FunctionProcesses can return a single value as output, and not a dictionary, so we should also return that
        if result and len(result) == 1 and self.SINGLE_OUTPUT_LINKNAME in result:
            return result[self.SINGLE_OUTPUT_LINKNAME]

        return result

    @override
    def _setup_db_record(self) -> None:
        """Set up the database record for the process."""
        super()._setup_db_record()
        self.node.store_source_info(self._func)

    @override
    def run(self) -> Optional['ExitCode']:
        """Run the process."""
        from .exit_code import ExitCode

        # The following conditional is required for the caching to properly work. Even if the source node has a process
        # state of `Finished` the cached process will still enter the running state. The process state will have then
        # been overridden by the engine to `Running` so we cannot check that, but if the `exit_status` is anything other
        # than `None`, it should mean this node was taken from the cache, so the process should not be rerun.
        if self.node.exit_status is not None:
            return self.node.exit_status

        # Split the inputs into positional and keyword arguments
        args = [None] * len(self._func_args)
        kwargs = {}

        for name, value in (self.inputs or {}).items():
            try:
                if self.spec().inputs[name].non_db:  # type: ignore[union-attr]
                    # Don't consider non-database inputs
                    continue
            except KeyError:
                pass  # No port found

            # Check if it is a positional arg, if not then keyword
            try:
                args[self._func_args.index(name)] = value
            except ValueError:
                kwargs[name] = value

        result = self._func(*args, **kwargs)

        if result is None or isinstance(result, ExitCode):
            return result

        if isinstance(result, Data):
            self.out(self.SINGLE_OUTPUT_LINKNAME, result)
        elif isinstance(result, collections.abc.Mapping):
            for name, value in result.items():
                self.out(name, value)
        else:
            raise TypeError(
                "Function process returned an output with unsupported type '{}'\n"
                'Must be a Data type or a mapping of {{string: Data}}'.format(result.__class__)
            )

        return ExitCode()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""A utility module with simple functions to format variables into strings for cli outputs."""


def format_relative_time(datetime):
    """
    Return a string formatted timedelta of the given datetime with respect to the current datetime

    :param datetime: the datetime to format
    :return: string representation of the relative time since the given datetime
    """
    from aiida.common import timezone
    from aiida.common.utils import str_timedelta

    timedelta = timezone.delta(datetime)

    return str_timedelta(timedelta, negative_to_zero=True, max_num_fields=1)


def format_state(process_state, paused=None, exit_status=None):
    """
    Return a string formatted representation of a process' state, which consists of its process state and exit status

    :param process_state: the process state
    :param exit_status: the process' exit status
    :return: string representation of the process' state
    """
    if process_state in ['excepted']:
        symbol = '\u2A2F'
    elif process_state in ['killed']:
        symbol = '\u2620'
    elif process_state in ['created', 'finished']:
        symbol = '\u23F9'
    elif process_state in ['running', 'waiting']:
        if paused is True:
            symbol = '\u23F8'
        else:
            symbol = '\u23F5'
    else:
        # Unknown process state, use invisible separator
        symbol = '\u00B7'  # middle dot

    if process_state == 'finished' and exit_status is not None:
        return f'{symbol} {format_process_state(process_state)} [{exit_status}]'

    return f'{symbol} {format_process_state(process_state)}'


def format_process_state(process_state):
    """
    Return a string formatted representation of the given process state

    :param process_state: the process state
    :return: string representation of process state
    """
    return f'{process_state.capitalize() if process_state else None}'


def format_sealed(sealed):
    """
    Return a string formatted representation of a node's sealed status

    :param sealed: the value for the sealed attribute of the node
    :return: string representation of seal status
    """
    return 'True' if sealed == 1 else 'False'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration from the "legacy" JSON format, to an sqlite database, and node uuid based repository to hash based."""
from contextlib import contextmanager
from datetime import datetime
from hashlib import sha256
from pathlib import Path, PurePosixPath
import shutil
import tarfile
from typing import Any, Callable, ContextManager, Dict, Iterator, List, Optional, Tuple, Union

from archive_path import ZipPath
from sqlalchemy import insert, select
from sqlalchemy.exc import IntegrityError

from aiida.common.exceptions import CorruptStorage, StorageMigrationError
from aiida.common.hashing import chunked_file_hash
from aiida.common.progress_reporter import get_progress_reporter
from aiida.repository.common import File, FileType
from aiida.storage.log import MIGRATE_LOGGER

from . import v1_db_schema as v1_schema
from ..utils import DB_FILENAME, REPO_FOLDER, create_sqla_engine
from .utils import update_metadata

_NODE_ENTITY_NAME = 'Node'
_GROUP_ENTITY_NAME = 'Group'
_COMPUTER_ENTITY_NAME = 'Computer'
_USER_ENTITY_NAME = 'User'
_LOG_ENTITY_NAME = 'Log'
_COMMENT_ENTITY_NAME = 'Comment'

file_fields_to_model_fields: Dict[str, Dict[str, str]] = {
    _NODE_ENTITY_NAME: {
        'dbcomputer': 'dbcomputer_id',
        'user': 'user_id'
    },
    _GROUP_ENTITY_NAME: {
        'user': 'user_id'
    },
    _COMPUTER_ENTITY_NAME: {},
    _LOG_ENTITY_NAME: {
        'dbnode': 'dbnode_id'
    },
    _COMMENT_ENTITY_NAME: {
        'dbnode': 'dbnode_id',
        'user': 'user_id'
    }
}

aiida_orm_to_backend = {
    _USER_ENTITY_NAME: v1_schema.DbUser,
    _GROUP_ENTITY_NAME: v1_schema.DbGroup,
    _NODE_ENTITY_NAME: v1_schema.DbNode,
    _COMMENT_ENTITY_NAME: v1_schema.DbComment,
    _COMPUTER_ENTITY_NAME: v1_schema.DbComputer,
    _LOG_ENTITY_NAME: v1_schema.DbLog,
}

LEGACY_TO_MAIN_REVISION = 'main_0000'


def perform_v1_migration(  # pylint: disable=too-many-locals
    inpath: Path,
    working: Path,
    new_zip: ZipPath,
    central_dir: Dict[str, Any],
    is_tar: bool,
    metadata: dict,
    data: dict,
) -> Path:
    """Perform the repository and JSON to SQLite migration.

    1. Iterate though the repository paths in the archive
    2. If a file, hash its contents and, if not already present, stream it to the new archive
    3. Store a mapping of the node UUIDs to a list of (path, hashkey or None if a directory) tuples

    :param inpath: the input path to the old archive
    :param metadata: the metadata to migrate
    :param data: the data to migrate

    :returns:the path to the sqlite database file
    """
    MIGRATE_LOGGER.report('Initialising new archive...')
    node_repos: Dict[str, List[Tuple[str, Optional[str]]]] = {}
    in_archive_context: Callable[[Path], ContextManager[Union[Path, ZipPath]]] = ZipPath
    if is_tar:
        # we cannot stream from a tar file performantly, so we extract it to disk first
        @contextmanager
        def _in_archive_context(_inpath):
            temp_folder = working / 'temp_unpack'
            with tarfile.open(_inpath, 'r') as tar:
                MIGRATE_LOGGER.report('Extracting tar archive...(may take a while)')
                tar.extractall(temp_folder)
            yield temp_folder
            MIGRATE_LOGGER.report('Removing extracted tar archive...')
            shutil.rmtree(temp_folder)

        in_archive_context = _in_archive_context

    with in_archive_context(inpath) as path:
        length = sum(1 for _ in path.glob('**/*'))
        base_parts = len(path.parts)
        with get_progress_reporter()(desc='Converting repo', total=length) as progress:
            for subpath in path.glob('**/*'):
                progress.update()
                parts = subpath.parts[base_parts:]
                # repository file are stored in the legacy archive as `nodes/uuid[0:2]/uuid[2:4]/uuid[4:]/path/...`
                if len(parts) < 6 or parts[0] != 'nodes' or parts[4] not in ('raw_input', 'path'):
                    continue
                uuid = ''.join(parts[1:4])
                posix_rel = PurePosixPath(*parts[5:])
                hashkey = None
                if subpath.is_file():
                    with subpath.open('rb') as handle:
                        hashkey = chunked_file_hash(handle, sha256)
                    if f'{REPO_FOLDER}/{hashkey}' not in central_dir:
                        (new_zip / f'{REPO_FOLDER}/{hashkey}').putfile(subpath)
                node_repos.setdefault(uuid, []).append((posix_rel.as_posix(), hashkey))
        MIGRATE_LOGGER.report(f'Unique repository files written: {len(central_dir)}')

    # convert the JSON database to SQLite
    _json_to_sqlite(working / DB_FILENAME, data, node_repos)

    # remove legacy keys from metadata and store
    metadata.pop('unique_identifiers', None)
    metadata.pop('all_fields_info', None)
    # remove legacy key nesting
    metadata['creation_parameters'] = metadata.pop('export_parameters', {})
    metadata['key_format'] = 'sha256'

    # update the version in the metadata
    update_metadata(metadata, LEGACY_TO_MAIN_REVISION)

    return working / DB_FILENAME


def _json_to_sqlite(  # pylint: disable=too-many-branches,too-many-locals
    outpath: Path, data: dict, node_repos: Dict[str, List[Tuple[str, Optional[str]]]], batch_size: int = 100
) -> None:
    """Convert a JSON archive format to SQLite."""
    from aiida.tools.archive.common import batch_iter

    MIGRATE_LOGGER.report('Converting DB to SQLite')

    engine = create_sqla_engine(outpath)
    v1_schema.ArchiveV1Base.metadata.create_all(engine)

    with engine.begin() as connection:
        # proceed in order of relationships
        for entity_type in (
            _USER_ENTITY_NAME, _COMPUTER_ENTITY_NAME, _GROUP_ENTITY_NAME, _NODE_ENTITY_NAME, _LOG_ENTITY_NAME,
            _COMMENT_ENTITY_NAME
        ):
            if not data['export_data'].get(entity_type, {}):
                continue
            length = len(data['export_data'].get(entity_type, {}))
            backend_cls = aiida_orm_to_backend[entity_type]
            with get_progress_reporter()(desc=f'Adding {entity_type}s', total=length) as progress:
                for nrows, rows in batch_iter(_iter_entity_fields(data, entity_type, node_repos), batch_size):
                    # to-do check for unused keys?
                    # to-do handle null values?
                    try:
                        connection.execute(insert(backend_cls.__table__), rows)  # type: ignore
                    except IntegrityError as exc:
                        raise StorageMigrationError(f'Database integrity error: {exc}') from exc
                    progress.update(nrows)

    if not (data['groups_uuid'] or data['links_uuid']):
        return None

    with engine.begin() as connection:

        # get mapping of node IDs to node UUIDs
        node_uuid_map = {
            uuid: pk for uuid, pk in connection.execute(select(v1_schema.DbNode.uuid, v1_schema.DbNode.id))  # pylint: disable=unnecessary-comprehension
        }

        # links
        if data['links_uuid']:

            def _transform_link(link_row):
                try:
                    input_id = node_uuid_map[link_row['input']]
                except KeyError:
                    raise StorageMigrationError(f'Database contains link with unknown input node: {link_row}')
                try:
                    output_id = node_uuid_map[link_row['output']]
                except KeyError:
                    raise StorageMigrationError(f'Database contains link with unknown output node: {link_row}')
                return {
                    'input_id': input_id,
                    'output_id': output_id,
                    'label': link_row['label'],
                    'type': link_row['type']
                }

            with get_progress_reporter()(desc='Adding Links', total=len(data['links_uuid'])) as progress:
                for nrows, rows in batch_iter(data['links_uuid'], batch_size, transform=_transform_link):
                    connection.execute(insert(v1_schema.DbLink.__table__), rows)
                    progress.update(nrows)

        # groups to nodes
        if data['groups_uuid']:
            # get mapping of node IDs to node UUIDs
            group_uuid_map = {
                uuid: pk for uuid, pk in connection.execute(select(v1_schema.DbGroup.uuid, v1_schema.DbGroup.id))  # pylint: disable=unnecessary-comprehension
            }
            length = sum(len(uuids) for uuids in data['groups_uuid'].values())
            unknown_nodes: Dict[str, set] = {}
            with get_progress_reporter()(desc='Adding Group-Nodes', total=length) as progress:
                for group_uuid, node_uuids in data['groups_uuid'].items():
                    group_id = group_uuid_map[group_uuid]
                    rows = []
                    for uuid in node_uuids:
                        if uuid in node_uuid_map:
                            rows.append({'dbnode_id': node_uuid_map[uuid], 'dbgroup_id': group_id})
                        else:
                            unknown_nodes.setdefault(group_uuid, set()).add(uuid)
                    connection.execute(insert(v1_schema.DbGroupNodes.__table__), rows)
                    progress.update(len(node_uuids))
            if unknown_nodes:
                MIGRATE_LOGGER.warning(f'Dropped unknown nodes in groups: {unknown_nodes}')


def _convert_datetime(key, value):
    if key in ('time', 'ctime', 'mtime') and value is not None:
        return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
    return value


def _iter_entity_fields(
    data,
    name: str,
    node_repos: Dict[str, List[Tuple[str, Optional[str]]]],
) -> Iterator[Dict[str, Any]]:
    """Iterate through entity fields."""
    keys = file_fields_to_model_fields.get(name, {})
    if name == _NODE_ENTITY_NAME:
        # here we merge in the attributes and extras before yielding
        attributes = data.get('node_attributes', {})
        extras = data.get('node_extras', {})
        for pk, all_fields in data['export_data'].get(name, {}).items():
            if pk not in attributes:
                raise CorruptStorage(f'Unable to find attributes info for Node with Pk={pk}')
            if pk not in extras:
                raise CorruptStorage(f'Unable to find extra info for Node with Pk={pk}')
            uuid = all_fields['uuid']
            repository_metadata = _create_repo_metadata(node_repos[uuid]) if uuid in node_repos else {}
            yield {
                **{keys.get(key, key): _convert_datetime(key, val) for key, val in all_fields.items()},
                **{
                    'id': pk,
                    'attributes': attributes[pk],
                    'extras': extras[pk],
                    'repository_metadata': repository_metadata
                }
            }
    else:
        for pk, all_fields in data['export_data'].get(name, {}).items():
            yield {**{keys.get(key, key): _convert_datetime(key, val) for key, val in all_fields.items()}, **{'id': pk}}


def _create_repo_metadata(paths: List[Tuple[str, Optional[str]]]) -> Dict[str, Any]:
    """Create the repository metadata.

    :param paths: list of (path, hashkey) tuples
    :return: the repository metadata
    """
    top_level = File()
    for _path, hashkey in paths:
        path = PurePosixPath(_path)
        if hashkey is None:
            _create_directory(top_level, path)
        else:
            directory = _create_directory(top_level, path.parent)
            directory.objects[path.name] = File(path.name, FileType.FILE, hashkey)
    return top_level.serialize()


def _create_directory(top_level: File, path: PurePosixPath) -> File:
    """Create a new directory with the given path.

    :param path: the relative path of the directory.
    :return: the created directory.
    """
    directory = top_level

    for part in path.parts:
        if part not in directory.objects:
            directory.objects[part] = File(part)

        directory = directory.objects[part]

    return directory



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Contain utility classes for "managers", i.e., classes that allow
to access members of other classes via TAB-completable attributes
(e.g. the class underlying `calculation.inputs` to allow to do `calculation.inputs.<label>`).
"""
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent, NotExistentAttributeError, NotExistentKeyError
from aiida.common.links import LinkType
from aiida.common.warnings import warn_deprecation

__all__ = ('NodeLinksManager', 'AttributeManager')


class NodeLinksManager:
    """
    A manager that allows to inspect, with tab-completion, nodes linked to a given one.
    See an example of its use in `CalculationNode.inputs`.
    """

    _namespace_separator = '__'

    def __init__(self, node, link_type, incoming):
        """
        Initialise the link manager.

        :param node: the reference node object
        :param link_type: the link_type to inspect
        :param incoming: if True, inspect incoming links, otherwise inspect
            outgoing links
        """
        # This import is here to avoid circular imports
        from aiida.orm import Node

        if not isinstance(node, Node):
            raise TypeError('node must be a valid AiiDA Node')
        self._node = node
        if not isinstance(link_type, LinkType):
            raise TypeError('link_type must be a valid LinkType')
        self._link_type = link_type
        self._incoming = incoming

    def _construct_attribute_dict(self, incoming):
        """Construct an attribute dict from all links of the node, recreating nested namespaces from flat link labels.

        :param incoming: if True, inspect incoming links, otherwise inspect outgoing links.
        """
        if incoming:
            links = self._node.base.links.get_incoming(link_type=self._link_type)
        else:
            links = self._node.base.links.get_outgoing(link_type=self._link_type)

        return AttributeDict(links.nested())

    def _get_keys(self):
        """Return the valid link labels, used e.g. to make getattr() work"""
        attribute_dict = self._construct_attribute_dict(self._incoming)

        return attribute_dict.keys()

    def _get_node_by_link_label(self, label):
        """Return the linked node with a given link label.

        Nested namespaces in link labels get represented by double underscores in the database. Up until now, the link
        manager didn't automatically unroll these again into nested namespaces and so a user was forced to pass the link
        with double underscores to dereference the corresponding node. For example, when used with the ``inputs``
        attribute of a ``ProcessNode`` one had to do:

            node.inputs.nested__sub__namespace

        Now it is possible to do

            node.inputs.nested.sub.namespace

        which is more intuitive since the double underscore replacement is just for the database and the user shouldn't
        even have to know about it. For compatibility we support the old version a bit longer and it will emit a
        deprecation warning.

        :param label: the link label connecting the current node to the node to get.
        """
        attribute_dict = self._construct_attribute_dict(self._incoming)
        try:
            node = attribute_dict[label]
        except KeyError as exception:
            # Check whether the label contains a double underscore, in which case we want to warn the user that this is
            # deprecated. However, we need to exclude labels that corresponds to dunder methods, i.e., those that start
            # and end with a double underscore.
            if (
                self._namespace_separator in label and
                not (label.startswith(self._namespace_separator) and label.endswith(self._namespace_separator))
            ):
                import functools
                warn_deprecation(
                    'dereferencing nodes with links containing double underscores is deprecated, simply replace '
                    'the double underscores with a single dot instead. For example: \n'
                    '`self.inputs.some__label` can be written as `self.inputs.some.label` instead.\n',
                    version=3
                )  # pylint: disable=no-member
                namespaces = label.split(self._namespace_separator)
                try:
                    return functools.reduce(lambda d, namespace: d.get(namespace), namespaces, attribute_dict)
                except TypeError as exc:
                    # This can be raised if part of the `namespaces` correspond to an actual leaf node, but is treated
                    # like a namespace
                    raise NotExistent from exc
                except AttributeError as exc:
                    # This will be raised if any of the intermediate namespaces don't exist, and so the label node does
                    # not exist.
                    raise NotExistent from exc
            raise NotExistent from exception

        return node

    def __dir__(self):
        """
        Allow to list all valid input links
        """
        node_attributes = self._get_keys()
        return sorted(set(list(dir(type(self))) + list(node_attributes)))

    def __iter__(self):
        node_attributes = self._get_keys()
        for k in node_attributes:
            yield k

    def __getattr__(self, name):
        """
        :param name: name of the attribute to be asked to the parser results.
        """
        try:
            return self._get_node_by_link_label(label=name)
        except NotExistent as exception:
            # Note: in order for TAB-completion to work, we need to raise an exception that also inherits from
            # `AttributeError`, so that `getattr(node.inputs, 'some_label', some_default)` returns `some_default`.
            # Otherwise, the exception is not caught by `getattr` and is propagated, instead of returning the default.
            prefix = 'input' if self._incoming else 'output'
            raise NotExistentAttributeError(
                f"Node<{self._node.pk}> does not have an {prefix} with link label '{name}'"
            ) from exception

    def __contains__(self, key):
        """Override the operator of the base class to emit deprecation warning if double underscore is used in key."""
        if self._namespace_separator in key:
            warn_deprecation(
                'The use of double underscores in keys is deprecated. Please expand the namespaces manually:\n'
                'instead of `nested__key in node.inputs`, use `nested in node.inputs and `key in node.inputs.nested`.',
                version=3
            )
            namespaces = key.split(self._namespace_separator)
            leaf = namespaces.pop()
            subdictionary = self
            for namespace in namespaces:
                try:
                    subdictionary = subdictionary[namespace]
                except KeyError:
                    return False
            return leaf in subdictionary

        return key in self._get_keys()

    def __getitem__(self, name):
        """
        interface to get to the parser results as a dictionary.

        :param name: name of the attribute to be asked to the parser results.
        """
        try:
            return self._get_node_by_link_label(label=name)
        except NotExistent as exception:
            # Note: in order for this class to behave as a dictionary, we raise an exception that also inherits from
            # `KeyError` - in this way, users can use the standard construct `try/except KeyError` and this will behave
            # like a standard dictionary.
            prefix = 'input' if self._incoming else 'output'
            raise NotExistentKeyError(
                f"Node<{self._node.pk}> does not have an {prefix} with link label '{name}'"
            ) from exception

    def __str__(self):
        """Return a string representation of the manager"""
        prefix = 'incoming' if self._incoming else 'outgoing'
        return f'Manager for {prefix} {self._link_type.value.upper()} links for node pk={self._node.pk}'

    def __repr__(self):
        return f'<{self.__class__.__name__}: {str(self)}>'


class AttributeManager:
    """
    An object used internally to return the attributes as a dictionary.
    This is currently used in :py:class:`~aiida.orm.nodes.data.dict.Dict`,
    for instance.

    :note: Important! It cannot be used to change variables, just to read
      them. To change values (of unstored nodes), use the proper Node methods.
    """

    def __init__(self, node):
        """
        :param node: the node object.
        """
        # Possibly add checks here
        # We cannot set `self._node` because it would go through the __setattr__ method
        # which uses said _node by calling `self._node.base.attributes.set(name, value)`.
        # Instead, we need to manually set it through the `self.__dict__` property.
        self.__dict__['_node'] = node

    def __dir__(self):
        """
        Allow to list the keys of the dictionary
        """
        return sorted(self._node.base.attributes.keys())

    def __iter__(self):
        """
        Return the keys as an iterator
        """
        for k in self._node.base.attributes.keys():
            yield k

    def _get_dict(self):
        """
        Return the internal dictionary
        """
        return dict(self._node.base.attributes.items())

    def __getattr__(self, name):
        """
        Interface to get to dictionary values, using the key as an attribute.

        :note: it works only for attributes that only contain letters, numbers
          and underscores, and do not start with a number.

        :param name: name of the key whose value is required.
        """
        return self._node.base.attributes.get(name)

    def __setattr__(self, name, value):
        self._node.base.attributes.set(name, value)

    def __getitem__(self, name):
        """
        Interface to get to dictionary values as a dictionary.

        :param name: name of the key whose value is required.
        """
        try:
            return self._node.base.attributes.get(name)
        except AttributeError as exception:
            raise KeyError(str(exception)) from exception



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to be used as a base for data containers that represent base python data types."""
from functools import singledispatch

from .data import Data

__all__ = ('BaseType', 'to_aiida_type')


@singledispatch
def to_aiida_type(value):
    """Turns basic Python types (str, int, float, bool) into the corresponding AiiDA types."""
    raise TypeError(f'Cannot convert value of type {type(value)} to AiiDA type.')


class BaseType(Data):
    """`Data` sub class to be used as a base for data containers that represent base python data types."""

    def __init__(self, value=None, **kwargs):
        try:
            getattr(self, '_type')
        except AttributeError:
            raise RuntimeError('Derived class must define the `_type` class member')

        super().__init__(**kwargs)

        self.value = value or self._type()  # pylint: disable=no-member

    @property
    def value(self):
        return self.base.attributes.get('value', None)

    @value.setter
    def value(self, value):
        self.base.attributes.set('value', self._type(value))  # pylint: disable=no-member

    def __str__(self):
        return f'{super().__str__()} value: {self.value}'

    def __eq__(self, other):
        if isinstance(other, BaseType):
            return self.value == other.value
        return self.value == other

    def new(self, value=None):
        return self.__class__(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Schema validation and migration utilities.

This code interacts directly with the database, outside of the ORM,
taking a `Profile` as input for the connection configuration.

.. important:: This code should only be accessed via the storage backend class, not directly!
"""
import contextlib
import os
import pathlib
from typing import ContextManager, Dict, Iterator, Optional

from alembic.command import downgrade, upgrade
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import MigrationContext, MigrationInfo
from alembic.script import ScriptDirectory
from disk_objectstore import Container
from sqlalchemy import String, Table, column, desc, insert, inspect, select, table
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.future.engine import Connection
from sqlalchemy.orm import Session

from aiida.common import exceptions
from aiida.manage.configuration.profile import Profile
from aiida.storage.log import MIGRATE_LOGGER
from aiida.storage.psql_dos.models.settings import DbSetting
from aiida.storage.psql_dos.utils import create_sqlalchemy_engine

TEMPLATE_LEGACY_DJANGO_SCHEMA = """
Database schema is using the legacy Django schema.
To migrate the database schema version to the current one, run the following command:

    verdi -p {profile_name} storage migrate
"""

TEMPLATE_INVALID_SCHEMA_VERSION = """
Database schema version `{schema_version_database}` is incompatible with the required schema version `{schema_version_code}`.
To migrate the database schema version to the current one, run the following command:

    verdi -p {profile_name} storage migrate
"""

ALEMBIC_REL_PATH = 'migrations'

REPOSITORY_UUID_KEY = 'repository|uuid'


class PsqlDostoreMigrator:
    """Class for validating and migrating `psql_dos` storage instances.

    .. important:: This class should only be accessed via the storage backend class (apart from for test purposes)
    """

    alembic_version_tbl_name = 'alembic_version'
    django_version_table = table(
        'django_migrations', column('id'), column('app', String(255)), column('name', String(255)), column('applied')
    )

    def __init__(self, profile: Profile) -> None:
        self.profile = profile

    @classmethod
    def get_schema_versions(cls) -> Dict[str, str]:
        """Return all available schema versions (oldest to latest).

        :return: schema version -> description
        """
        return {entry.revision: entry.doc for entry in reversed(list(cls._alembic_script().walk_revisions()))}

    @classmethod
    def get_schema_version_head(cls) -> str:
        """Return the head schema version for this storage, i.e. the latest schema this storage can be migrated to."""
        return cls._alembic_script().revision_map.get_current_head('main')

    def _connection_context(self, connection: Optional[Connection] = None) -> ContextManager[Connection]:
        """Return a context manager, with a connection to the database.

        :raises: `UnreachableStorage` if the database connection fails
        """
        if connection is not None:
            return contextlib.nullcontext(connection)
        try:
            return create_sqlalchemy_engine(self.profile.storage_config).connect()
        except OperationalError as exception:
            raise exceptions.UnreachableStorage(f'Could not connect to database: {exception}') from exception

    def get_schema_version_profile(self, _connection: Optional[Connection] = None, check_legacy=False) -> Optional[str]:
        """Return the schema version of the backend instance for this profile.

        Note, the version will be None if the database is empty or is a legacy django database.
        """
        with self._migration_context(_connection) as context:
            version = context.get_current_revision()
        if version is None and check_legacy:
            with self._connection_context(_connection) as connection:
                stmt = select(self.django_version_table.c.name).where(self.django_version_table.c.app == 'db')
                stmt = stmt.order_by(desc(self.django_version_table.c.id)).limit(1)
                try:
                    return connection.execute(stmt).scalar()
                except (OperationalError, ProgrammingError):
                    connection.rollback()
        return version

    def validate_storage(self) -> None:
        """Validate that the storage for this profile

        1. That the database schema is at the head version, i.e. is compatible with the code API.
        2. That the repository ID is equal to the UUID set in the database

        :raises: :class:`aiida.common.exceptions.UnreachableStorage` if the storage cannot be connected to
        :raises: :class:`aiida.common.exceptions.IncompatibleStorageSchema`
            if the storage is not compatible with the code API.
        :raises: :class:`aiida.common.exceptions.CorruptStorage`
            if the repository ID is not equal to the UUID set in thedatabase.
        """
        with self._connection_context() as connection:

            # check there is an alembic_version table from which to get the schema version
            if not inspect(connection).has_table(self.alembic_version_tbl_name):
                # if not present, it might be that this is a legacy django database
                if inspect(connection).has_table(self.django_version_table.name):
                    raise exceptions.IncompatibleStorageSchema(
                        TEMPLATE_LEGACY_DJANGO_SCHEMA.format(profile_name=self.profile.name)
                    )
                raise exceptions.IncompatibleStorageSchema('The database has no known version.')

            # now we can check that the alembic version is the latest
            schema_version_code = self.get_schema_version_head()
            schema_version_database = self.get_schema_version_profile(connection, check_legacy=False)
            if schema_version_database != schema_version_code:
                raise exceptions.IncompatibleStorageSchema(
                    TEMPLATE_INVALID_SCHEMA_VERSION.format(
                        schema_version_database=schema_version_database,
                        schema_version_code=schema_version_code,
                        profile_name=self.profile.name
                    )
                )

            # check that we can access the disk-objectstore container, and get its id
            filepath = pathlib.Path(self.profile.repository_path) / 'container'
            container = Container(filepath)
            try:
                container_id = container.container_id
            except Exception as exc:
                raise exceptions.UnreachableStorage(f'Could not access disk-objectstore {filepath}: {exc}') from exc

            # finally, we check that the ID set within the disk-objectstore is equal to the one saved in the database,
            # i.e. this container is indeed the one associated with the db
            stmt = select(DbSetting.val).where(DbSetting.key == REPOSITORY_UUID_KEY)
            repo_uuid = connection.execute(stmt).scalar_one_or_none()
            if repo_uuid is None:
                raise exceptions.CorruptStorage('The database has no repository UUID set.')
            if repo_uuid != container_id:
                raise exceptions.CorruptStorage(
                    f'The database has a repository UUID configured to {repo_uuid} '
                    f'but the disk-objectstore\'s is {container_id}.'
                )

    def initialise(self) -> None:
        """Generate the initial storage schema for this profile, from the ORM models."""
        from aiida.storage.psql_dos.backend import CONTAINER_DEFAULTS
        from aiida.storage.psql_dos.models.base import get_orm_metadata

        # setup the database
        # see: https://alembic.sqlalchemy.org/en/latest/cookbook.html#building-an-up-to-date-database-from-scratch
        get_orm_metadata().create_all(create_sqlalchemy_engine(self.profile.storage_config))

        # setup the repository
        filepath = pathlib.Path(self.profile.repository_path) / 'container'
        container = Container(filepath)
        container.init_container(clear=True, **CONTAINER_DEFAULTS)

        with create_sqlalchemy_engine(self.profile.storage_config).begin() as conn:
            # Create a "sync" between the database and repository, by saving its UUID in the settings table
            # this allows us to validate inconsistencies between the two
            conn.execute(
                insert(DbSetting
                       ).values(key=REPOSITORY_UUID_KEY, val=container.container_id, description='Repository UUID')
            )

            # finally, generate the version table, "stamping" it with the most recent revision
            with self._migration_context(conn) as context:
                context.stamp(context.script, 'main@head')

    def migrate(self) -> None:
        """Migrate the storage for this profile to the head version.

        :raises: :class:`~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        """
        # the database can be in one of a few states:
        # 1. Completely empty -> we can simply initialise it with the current ORM schema
        # 2. Legacy django database -> we transfer the version to alembic, migrate to the head of the django branch,
        #    reset the revision as one on the main branch, and then migrate to the head of the main branch
        # 3. Legacy sqlalchemy database -> we migrate to the head of the sqlalchemy branch,
        #    reset the revision as one on the main branch, and then migrate to the head of the main branch
        # 4. Already on the main branch -> we migrate to the head of the main branch

        with self._connection_context() as connection:
            if not inspect(connection).has_table(self.alembic_version_tbl_name):
                if not inspect(connection).has_table(self.django_version_table.name):
                    # the database is assumed to be empty, so we need to initialise it
                    MIGRATE_LOGGER.report('initialising empty storage schema')
                    self.initialise()
                    return
                # the database is a legacy django one,
                # so we need to copy the version from the 'django_migrations' table to the 'alembic_version' one
                legacy_version = self.get_schema_version_profile(connection, check_legacy=True)
                if legacy_version is None:
                    raise exceptions.StorageMigrationError(
                        'No schema version could be read from the database. '
                        "Check that either the 'alembic_version' or 'django_migrations' tables "
                        'are present and accessible, using e.g. `verdi devel run-sql "SELECT * FROM alembic_version"`'
                    )
                # the version should be of the format '00XX_description'
                version = f'django_{legacy_version[:4]}'
                with self._migration_context(connection) as context:
                    context.stamp(context.script, version)
                connection.commit()
                # now we can continue with the migration as normal
            else:
                version = self.get_schema_version_profile(connection)

        # find what branch the current version is on
        branches = self._alembic_script().revision_map.get_revision(version).branch_labels

        if 'django' in branches or 'sqlalchemy' in branches:
            # migrate up to the top of the respective legacy branches
            if 'django' in branches:
                MIGRATE_LOGGER.report('Migrating to the head of the legacy django branch')
                self.migrate_up('django@head')
            elif 'sqlalchemy' in branches:
                MIGRATE_LOGGER.report('Migrating to the head of the legacy sqlalchemy branch')
                self.migrate_up('sqlalchemy@head')
            # now re-stamp with the comparable revision on the main branch
            with self._connection_context() as connection:
                with self._migration_context(connection) as context:
                    context._ensure_version_table(purge=True)  # pylint: disable=protected-access
                    context.stamp(context.script, 'main_0001')
                connection.commit()

        # finally migrate to the main head revision
        MIGRATE_LOGGER.report('Migrating to the head of the main branch')
        self.migrate_up('main@head')

    def migrate_up(self, version: str) -> None:
        """Migrate the database up to a specific version.

        :param version: string with schema version to migrate to
        """
        with self._alembic_connect() as config:
            upgrade(config, version)

    def migrate_down(self, version: str) -> None:
        """Migrate the database down to a specific version.

        :param version: string with schema version to migrate to
        """
        with self._alembic_connect() as config:
            downgrade(config, version)

    @staticmethod
    def _alembic_config():
        """Return an instance of an Alembic `Config`."""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        config = Config()
        config.set_main_option('script_location', os.path.join(dir_path, ALEMBIC_REL_PATH))
        return config

    @classmethod
    def _alembic_script(cls):
        """Return an instance of an Alembic `ScriptDirectory`."""
        return ScriptDirectory.from_config(cls._alembic_config())

    @contextlib.contextmanager
    def _alembic_connect(self, _connection: Optional[Connection] = None) -> Iterator[Config]:
        """Context manager to return an instance of an Alembic configuration.

        The profiles's database connection is added in the `attributes` property, through which it can then also be
        retrieved, also in the `env.py` file, which is run when the database is migrated.
        """
        with self._connection_context(_connection) as connection:
            config = self._alembic_config()
            config.attributes['connection'] = connection  # pylint: disable=unsupported-assignment-operation
            config.attributes['aiida_profile'] = self.profile  # pylint: disable=unsupported-assignment-operation

            def _callback(step: MigrationInfo, **kwargs):  # pylint: disable=unused-argument
                """Callback to be called after a migration step is executed."""
                from_rev = step.down_revision_ids[0] if step.down_revision_ids else '<base>'
                MIGRATE_LOGGER.report(f'- {from_rev} -> {step.up_revision_id}')

            config.attributes['on_version_apply'] = _callback  # pylint: disable=unsupported-assignment-operation

            yield config

    @contextlib.contextmanager
    def _migration_context(self, _connection: Optional[Connection] = None) -> Iterator[MigrationContext]:
        """Context manager to return an instance of an Alembic migration context.

        This migration context will have been configured with the current database connection, which allows this context
        to be used to inspect the contents of the database, such as the current revision.
        """
        with self._alembic_connect(_connection) as config:
            script = ScriptDirectory.from_config(config)
            with EnvironmentContext(config, script) as context:
                context.configure(context.config.attributes['connection'])
                yield context.get_context()

    # the following are used for migration tests

    @contextlib.contextmanager
    def session(self) -> Iterator[Session]:
        """Context manager to return a session for the database."""
        with self._connection_context() as connection:
            session = Session(connection.engine, future=True)
            try:
                yield session
            except Exception:
                session.rollback()
                raise
            finally:
                session.close()

    def get_current_table(self, table_name: str) -> Table:
        """Return a table instantiated at the correct migration.

        Note that this is obtained by inspecting the database and not by looking into the models file.
        So, special methods possibly defined in the models files/classes are not present.
        """
        with self._connection_context() as connection:
            base = automap_base()
            base.prepare(autoload_with=connection.engine)
            return getattr(base.classes, table_name)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for `WorkChain` implementations."""
from functools import partial
from inspect import getfullargspec
from types import FunctionType  # pylint: disable=no-name-in-module
from typing import List, NamedTuple, Optional, Union

from wrapt import decorator

from ..exit_code import ExitCode

__all__ = ('ProcessHandlerReport', 'process_handler')


class ProcessHandlerReport(NamedTuple):
    """A namedtuple to define a process handler report for a :class:`aiida.engine.BaseRestartWorkChain`.

    This namedtuple should be returned by a process handler of a work chain instance if the condition of the handler was
    met by the completed process. If no further handling should be performed after this method the `do_break` field
    should be set to `True`.
    If the handler encountered a fatal error and the work chain needs to be terminated, an `ExitCode` with
    non-zero exit status can be set. This exit code is what will be set on the work chain itself. This works because the
    value of the `exit_code` field returned by the handler, will in turn be returned by the `inspect_process` step and
    returning a non-zero exit code from any work chain step will instruct the engine to abort the work chain.

    :param do_break: boolean, set to `True` if no further process handlers should be called, default is `False`
    :param exit_code: an instance of the :class:`~aiida.engine.processes.exit_code.ExitCode` tuple.
        If not explicitly set, the default `ExitCode` will be instantiated,
        which has status `0` meaning that the work chain step will be considered
        successful and the work chain will continue to the next step.
    """
    do_break: bool = False
    exit_code: ExitCode = ExitCode()


def process_handler(
    wrapped: Optional[FunctionType] = None,
    *,
    priority: int = 0,
    exit_codes: Union[None, ExitCode, List[ExitCode]] = None,
    enabled: bool = True
) -> FunctionType:
    """Decorator to register a :class:`~aiida.engine.BaseRestartWorkChain` instance method as a process handler.

    The decorator will validate the `priority` and `exit_codes` optional keyword arguments and then add itself as an
    attribute to the `wrapped` instance method. This is used in the `inspect_process` to return all instance methods of
    the class that have been decorated by this function and therefore are considered to be process handlers.

    Requirements on the function signature of process handling functions. The function to which the decorator is applied
    needs to take two arguments:

        * `self`: This is the instance of the work chain itself
        * `node`: This is the process node that finished and is to be investigated

    The function body typically consists of a conditional that will check for a particular problem that might have
    occurred for the sub process. If a particular problem is handled, the process handler should return an instance of
    the :class:`aiida.engine.ProcessHandlerReport` tuple. If no other process handlers should be considered, the set
    `do_break` attribute should be set to `True`. If the work chain is to be aborted entirely, the `exit_code` of the
    report can be set to an `ExitCode` instance with a non-zero status.

    :param wrapped: the work chain method to register the process handler with
    :param priority: optional integer that defines the order in which registered handlers will be called during the
        handling of a finished process. Higher priorities will be handled first. Default value is `0`. Multiple handlers
        with the same priority is allowed, but the order of those is not well defined.
    :param exit_codes: single or list of `ExitCode` instances. If defined, the handler will return `None` if the exit
        code set on the `node` does not appear in the `exit_codes`. This is useful to have a handler called only when
        the process failed with a specific exit code.
    :param enabled: boolean, by default True, which will cause the handler to be called during `inspect_process`. When
        set to `False`, the handler will be skipped. This static value can be overridden on a per work chain instance
        basis through the input `handler_overrides`.
    """
    if wrapped is None:
        return partial(
            process_handler, priority=priority, exit_codes=exit_codes, enabled=enabled
        )  # type: ignore[return-value]

    if not isinstance(wrapped, FunctionType):
        raise TypeError('first argument can only be an instance method, use keywords for decorator arguments.')

    if not isinstance(priority, int):
        raise TypeError('the `priority` keyword should be an integer.')

    if exit_codes is not None and not isinstance(exit_codes, list):
        exit_codes = [exit_codes]

    if exit_codes and any(not isinstance(exit_code, ExitCode) for exit_code in exit_codes):
        raise TypeError('`exit_codes` keyword should be an instance of `ExitCode` or list thereof.')

    if not isinstance(enabled, bool):
        raise TypeError('the `enabled` keyword should be a boolean.')

    handler_args = getfullargspec(wrapped)[0]

    if len(handler_args) != 2:
        raise TypeError(f'process handler `{wrapped.__name__}` has invalid signature: should be (self, node)')

    wrapped.decorator = process_handler  # type: ignore[attr-defined]
    wrapped.priority = priority  # type: ignore[attr-defined]
    wrapped.enabled = enabled  # type: ignore[attr-defined]

    @decorator
    def wrapper(wrapped, instance, args, kwargs):

        # When the handler will be called by the `BaseRestartWorkChain` it will pass the node as the only argument
        node = args[0]

        if exit_codes is not None and node.exit_status not in [
            exit_code.status for exit_code in exit_codes  # type: ignore[union-attr]
        ]:
            result = None
        else:
            result = wrapped(*args, **kwargs)

        # Append the name and return value of the current process handler to the `considered_handlers` extra.
        try:
            considered_handlers = instance.node.base.extras.get(instance._considered_handlers_extra, [])  # pylint: disable=protected-access
            current_process = considered_handlers[-1]
        except IndexError:
            # The extra was never initialized, so we skip this functionality
            pass
        else:
            # Append the name of the handler to the last list in `considered_handlers` and save it
            serialized = result
            if isinstance(serialized, ProcessHandlerReport):
                serialized = {'do_break': serialized.do_break, 'exit_status': serialized.exit_code.status}
            current_process.append((wrapped.__name__, serialized))
            instance.node.base.extras.set(instance._considered_handlers_extra, considered_handlers)  # pylint: disable=protected-access

        return result

    return wrapper(wrapped)  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Common password and hash generation functions."""
from collections import OrderedDict, abc
from datetime import date, datetime, timezone
from decimal import Decimal
from functools import singledispatch
import hashlib
from itertools import chain
import numbers
from operator import itemgetter
import secrets
import string
import typing
import uuid

from aiida.common.constants import AIIDA_FLOAT_PRECISION
from aiida.common.exceptions import HashingError
from aiida.common.utils import DatetimePrecision

from .folders import Folder


def get_random_string(length: int = 12) -> str:
    """Return a securely generated random string.

    The default length of 12 with the all ASCII letters and digits returns a 71-bit value:

        log_2((26+26+10)^12) =~ 71 bits

    :param length: The number of characters to use for the string.
    """
    alphabet = string.ascii_letters + string.digits
    return ''.join(secrets.choice(alphabet) for i in range(length))


BLAKE2B_OPTIONS = {
    'fanout': 0,  # unlimited fanout/depth mode
    'depth': 2,  # has fixed depth of 2
    'digest_size': 32,  # we do not need a cryptographically relevant digest
    'inner_size': 64,  # ... but still use 64 as the inner size
}


def chunked_file_hash(
    handle: typing.BinaryIO, hash_cls: typing.Any, chunksize: int = 524288, **kwargs: typing.Any
) -> str:
    """Return the hash for the given file handle

    Will read the file in chunks, which should be opened in 'rb' mode.

    :param handle: a file handle, opened in 'rb' mode.
    :param hash_cls: a class implementing hashlib._Hash
    :param chunksize: number of bytes to chunk the file read in
    :param kwargs: arguments to pass to the hasher initialisation
    :return: the hash hexdigest (the hash key)
    """
    hasher = hash_cls(**kwargs)
    while True:
        chunk = handle.read(chunksize)
        hasher.update(chunk)

        if not chunk:
            # Empty returned value: EOF
            break

    return hasher.hexdigest()


def make_hash(object_to_hash, **kwargs):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable or nonhashable types (including lists, tuples, sets, and
    dictionaries).

    :param object_to_hash: the object to hash

    :returns: a unique hash

    There are a lot of modules providing functionalities to create unique
    hashes for hashable values.
    However, getting hashes for nonhashable items like sets or dictionaries is
    not easily doable because order is not fixed.
    This leads to the peril of getting different hashes for the same
    dictionary.

    This function avoids this by recursing through nonhashable items and
    hashing iteratively. Uses python's sorted function to sort unsorted
    sets and dictionaries by sorting the hashed keys.
    """
    hashes = _make_hash(object_to_hash, **kwargs)  # pylint: disable=assignment-from-no-return

    # use the Unlimited fanout hashing protocol outlined in
    #   https://blake2.net/blake2_20130129.pdf
    final_hash = hashlib.blake2b(node_depth=1, last_node=True, **BLAKE2B_OPTIONS)

    for sub in hashes:
        final_hash.update(sub)

    # add an empty last leaf node
    final_hash.update(hashlib.blake2b(node_depth=0, last_node=True, **BLAKE2B_OPTIONS).digest())

    return final_hash.hexdigest()


@singledispatch
def _make_hash(object_to_hash, **_):
    """
    Implementation of the ``make_hash`` function. The hash is created as a
    28 byte integer, and only later converted to a string.
    """
    raise HashingError(f'Value of type {type(object_to_hash)} cannot be hashed')


def _single_digest(obj_type, obj_bytes=b''):
    return hashlib.blake2b(obj_bytes, person=obj_type.encode('ascii'), node_depth=0, **BLAKE2B_OPTIONS).digest()


_END_DIGEST = _single_digest(')')


@_make_hash.register(bytes)
def _(bytes_obj, **kwargs):
    """Hash arbitrary byte strings."""
    return [_single_digest('str', bytes_obj)]


@_make_hash.register(str)
def _(val, **kwargs):
    """Convert strings explicitly to bytes."""
    return [_single_digest('str', val.encode('utf-8'))]


@_make_hash.register(abc.Sequence)
def _(sequence_obj, **kwargs):
    # unpack the list and use the elements
    return [_single_digest('list(')] + list(chain.from_iterable(_make_hash(i, **kwargs) for i in sequence_obj)
                                            ) + [_END_DIGEST]


@_make_hash.register(abc.Set)
def _(set_obj, **kwargs):
    # turn the set objects into a list of hashes which are always sortable,
    # then return a flattened list of the hashes
    return [_single_digest('set(')] + list(chain.from_iterable(sorted(_make_hash(i, **kwargs) for i in set_obj))
                                           ) + [_END_DIGEST]


@_make_hash.register(abc.Mapping)
def _(mapping, **kwargs):
    """Hashing arbitrary mapping containers (dict, OrderedDict) by first sorting by hashed keys"""

    def hashed_key_mapping():
        for key, value in mapping.items():
            yield (_make_hash(key, **kwargs), value)

    return [_single_digest('dict(')] + list(
        chain.from_iterable(
            (k_digest + _make_hash(val, **kwargs)) for k_digest, val in sorted(hashed_key_mapping(), key=itemgetter(0))
        )
    ) + [_END_DIGEST]


@_make_hash.register(OrderedDict)
def _(mapping, **kwargs):
    """
    Hashing of OrderedDicts

    :param odict_as_unordered: hash OrderedDicts as normal dicts (mostly for testing)
    """

    if kwargs.get('odict_as_unordered', False):
        return _make_hash.registry[abc.Mapping](mapping)

    return ([_single_digest('odict(')] + list(
        chain.from_iterable((_make_hash(key, **kwargs) + _make_hash(val, **kwargs)) for key, val in mapping.items())
    ) + [_END_DIGEST])


@_make_hash.register(numbers.Real)
def _(val, **kwargs):
    """
    Before hashing a float, convert to a string (via rounding) and with a fixed number of digits after the comma.
    Note that the `_single_digest` requires a bytes object so we need to encode the utf-8 string first
    """
    return [_single_digest('float', float_to_text(val, sig=AIIDA_FLOAT_PRECISION).encode('utf-8'))]


@_make_hash.register(Decimal)
def _(val, **kwargs):
    """
    While a decimal can be converted exactly to a string which captures all characteristics of the underlying
    implementation, we also need compatibility with "equal" representations as int or float. Hence we are checking
    for the exponent (which is negative if there is a fractional component, 0 otherwise) and get the same hash
    as for a corresponding float or int.
    """
    if val.as_tuple().exponent < 0:
        return [_single_digest('float', float_to_text(val, sig=AIIDA_FLOAT_PRECISION).encode('utf-8'))]
    return [_single_digest('int', f'{val}'.encode('utf-8'))]


@_make_hash.register(numbers.Complex)
def _(val, **kwargs):
    """
    In case of a complex number, use the same encoding of two floats and join them with a special symbol (a ! here).
    """
    return [
        _single_digest(
            'complex', '{}!{}'.format(
                float_to_text(val.real, sig=AIIDA_FLOAT_PRECISION), float_to_text(val.imag, sig=AIIDA_FLOAT_PRECISION)
            ).encode('utf-8')
        )
    ]


@_make_hash.register(numbers.Integral)
def _(val, **kwargs):
    """get the hash of the little-endian signed long long representation of the integer"""
    return [_single_digest('int', f'{val}'.encode('utf-8'))]


@_make_hash.register(bool)
def _(val, **kwargs):
    return [_single_digest('bool', b'\x01' if val else b'\x00')]


@_make_hash.register(type(None))
def _(val, **kwargs):
    return [_single_digest('none')]


@_make_hash.register(datetime)
def _(val, **kwargs):
    """hashes the little-endian rep of the float <epoch-seconds>.<subseconds>"""
    # see also https://stackoverflow.com/a/8778548 for an excellent elaboration
    if val.tzinfo is None or val.utcoffset() is None:
        val = val.replace(tzinfo=timezone.utc)

    timestamp = val.timestamp()
    return [_single_digest('datetime', float_to_text(timestamp, sig=AIIDA_FLOAT_PRECISION).encode('utf-8'))]


@_make_hash.register(date)
def _(val, **kwargs):
    """Hashes the string representation in ISO format of the `datetime.date` object."""
    return [_single_digest('date', val.isoformat().encode('utf-8'))]


@_make_hash.register(uuid.UUID)
def _(val, **kwargs):
    return [_single_digest('uuid', val.bytes)]


@_make_hash.register(DatetimePrecision)
def _(datetime_precision, **kwargs):
    """ Hashes for DatetimePrecision object
    """
    return [_single_digest('dt_prec')] + list(
        chain.from_iterable(_make_hash(i, **kwargs) for i in [datetime_precision.dtobj, datetime_precision.precision])
    ) + [_END_DIGEST]


@_make_hash.register(Folder)
def _(folder, **kwargs):
    """
    Hash the content of a Folder object. The name of the folder itself is actually ignored
    :param ignored_folder_content: list of filenames to be ignored for the hashing
    """

    ignored_folder_content = kwargs.get('ignored_folder_content', [])

    def folder_digests(subfolder):
        """traverses the given folder and yields digests for the contained objects"""
        for name, isfile in sorted(subfolder.get_content_list(only_paths=False), key=itemgetter(0)):
            if name in ignored_folder_content:
                continue

            if isfile:
                yield _single_digest('fname', name.encode('utf-8'))
                with subfolder.open(name, mode='rb') as fhandle:
                    yield _single_digest('fcontent', fhandle.read())
            else:
                yield _single_digest('dir(', name.encode('utf-8'))
                for digest in folder_digests(subfolder.get_subfolder(name)):
                    yield digest
                yield _END_DIGEST

    return [_single_digest('folder')] + list(folder_digests(folder))


def float_to_text(value, sig):
    """
    Convert float to text string for computing hash.
    Preseve up to N significant number given by sig.

    :param value: the float value to convert
    :param sig: choose how many digits after the comma should be output
    """
    if value == 0:
        value = 0.  # Identify value of -0. and overwrite with 0.
    fmt = f'{{:.{sig}g}}'
    return fmt.format(value)
